#ifndef CLASS_ALIGN5_ICC
#define CLASS_ALIGN5_ICC


// implementation of the member functions for ALIGN5 class library
// ===================================================================
// Author: Martin Braeuer 
//
// NOTE: The name is derrived from the alignment methods:
//       ALIGN 1 => Target-spot pre-tracking alignment
//       ALIGN 2 => pre tracking alignment using seed/ 
//                  confirmation combis
//       ALIGN 3 => After tracking alignment, using 
//                  accepted hits, having seed/confirmation 
//                  combis==ResidualElements==REle's, 
//                  lateral alignment
//       ALIGN 4 => As No. 3, having rotational and longi-
//                  tudinal degrees of freedom
//       ALIGN 5 => Storing of the REle's in matrices
//
// Modification history:
// MBr  27.05.99: Porting from align5.cc V2.0 to class
// MBr  20.08.99: bug with update fixed.. fChange_GEDE() applied to main..
// MBr  14.09.99: Porting class to VDSLIB 2-19 
//                    => align5c1-0
// MBr  30.10.99: unsused vars removed, acess to matrix 
//                and vector, better printing
//                    => align5c1-1
// MBr  31.10.99: fill the matrix with user routines
//                and vector, better printing
//                    => align5c1-2
// MBr  17.11.99: improved par-order, bad-chip-side mechanism
//                    =>
// MBr  30.11.99: bad-chip-side mechanism to other Class
//                    => align5c1-3
// AG   11.05.00: use ARTE_exit() instead of exit()
// ==================================================================

// includes :
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <math.h>
#include  <malloc.h>

// STL stuff:
#include "map.h"
#include "multimap.h"

#include  "grr.h"
#include  "iol.h"
#include  "hbook.h"
#include  "marvin.h"
#include  "vdsleda.h"

// VDS specific:
#include "Struct_Gede.h"   // use Gede_t structs
#include "Struct_Seed.h"   // use Seed_t structs
#include "Struct_Hitb.h"   // use Hitb_t structs
#include "Struct_Track.h"  // use Track_t structs

#include "Class_GEDE.h"    // geometry information
#include "Class_SEED.h"    // seed generation
#include "Class_HOLMES.h"  // holmes track finding

// include the class's header file:
#include "alca/AC_SVD_ALIGN5.h"
#include "farm/farm.hh"

// Test of better equality of alignment parameters:
#define ZSFACT  1.0
#define RSFACT  1.0

// Contstants for fit systems :
#define EPSILON    3.0e-5 // small but significant num for fit'ers, NOT USED
#define T_SMALL    9.0e-8 // smallest significant number of COMP
#define T_SINGULAR 1.0e-11 // biggest SV*TSINGULAR == smallest non sing value

/* ==================================================================
 * == ALIGN5_class constructor : std constructor                   ==
 * == needs : ---                                                  ==
 * == Result : Failure message because of missing HOLMES pointer   ==
 * ==                                                              ==
 * ================================================================== 
 */
ALIGN5_class::ALIGN5_class() {
  fError_out("Default constructor call failed!");
  return;
}

/* ==================================================================
 * == ALIGN5_class constructor : Prepare structures for align5     ==
 * == Parameter: HOLMES Pointer                                    ==
 * ==            lateral, axial, rotational flag (1==align)        ==
 * ==                                                              ==
 * ================================================================== 
 */
ALIGN5_class::ALIGN5_class(HOLMES_class *HOLMESp, 
			   int lat, int axi, int rot,int noalign[],
			   int (*p1)(int, int, int, double  ),
			   int (*p2)(int, int, int, double *) ) {
  prepare = 0; // This 'makes' a new class -> NOTHING can be prepared!
  if (fPrepare(HOLMESp,lat,axi,rot,noalign,p1,p2) == 1) return;
  fError_out("Constructor call failed!");
  return;
}

/* ==================================================================
 * == ALIGN5_class destructor : Delete structures for align5       ==
 * ==                                                              ==
 * ================================================================== 
 */
ALIGN5_class::~ALIGN5_class() {
 fClear();
}

/* ==================================================================
 * == fClear : Clear/unbook all allocated structures               ==
 * ==                                                              ==
 * ================================================================== 
 */
void ALIGN5_class::fClear() {
  if (prepare>0) { // only if prepared crrectly..
    delete []pars; pars = NULL;
    if (svec     !=NULL ) { free((char*) svec     );svec       = NULL;}
    if (avec     !=NULL ) { free((char*) avec     );avec       = NULL;}
    if (esvec    !=NULL ) { free((char*) esvec    );esvec      = NULL;}
    if (amat     !=NULL ) { free((char*) amat     );amat       = NULL;}
    if (atam     !=NULL ) { free((char*) atam     );atam       = NULL;}
    if (avsing   !=NULL ) { free((char*) avsing   );avsing     = NULL;}
    if (mp       !=NULL ) { free((char*) mp       );mp         = NULL;}
    if (hiscount !=NULL ) { free((char*) hiscount );hiscount   = NULL;}
    if (gede     !=NULL ) { free((char*) gede     );gede       = NULL;}
    if (type     !=NULL ) { free((char*) type     );type       = NULL;}
    prepare=0;
    gather =0;
    solve  =0;
    npars  =0;
    sum0   =double(0);  
    dgf    =0; 
    nreles =0;
    ntrack =0;
  }
  return;
}



/* ==================================================================
 * == fGetPars : Get number of parameters                          ==
 * == needs : prepare has to run before (0 value otherwise)        ==
 * ================================================================== 
 */
int ALIGN5_class::fGetPars() {
  if (prepare != 0) return(npars);
  return(0);
}

/* ==================================================================
 * == fGetMatrix : Get the pars x pars Matrix                      ==
 * == needs : Gather has to run before (0 value otherwise)         ==
 * ================================================================== 
 */
double * ALIGN5_class::fGetMatrix() {
  if (gather != 0) return(amat);
  return(0);
}

/* ==================================================================
 * == fGetVector : Get the pars size Vector                        ==
 * == needs : Gather has to run before (0 value otherwise)         ==
 * ================================================================== 
 */
double * ALIGN5_class::fGetVector() {
  if (gather != 0) return(avec);
  return(0);
}

/* ==================================================================
 * == fGetChanges : Get the pars size Vector with solution         ==
 * == needs : Solve has to run before (0 value otherwise)          ==
 * ================================================================== 
 */
double * ALIGN5_class::fGetChanges() {
  if (solve != 0) return(svec);
  return(0);
}

/* ==================================================================
 * == fGetTracks : Get number of processed Tracks                  ==
 * == needs : Gather has to run before (0 value otherwise)         ==
 * ================================================================== 
 */
int ALIGN5_class::fGetTracks() {
  if (gather != 0) return(ntrack);
  return(0);
}

/* ==================================================================
 * == fGetReles : Get number of Residual elements                  ==
 * == needs : Gather has to run before (0 value otherwise)         ==
 * ================================================================== 
 */
int ALIGN5_class::fGetReles() {
  if (gather != 0) return(nreles);
  return(0);
}

/* ==================================================================
 * == fGetDgf : Get number of degrees of freedom                   ==
 * == needs : Gather has to run before (0 value otherwise)         ==
 * ================================================================== 
 */
int ALIGN5_class::fGetDgf() {
  if (gather != 0) return(dgf);
  return(0);
}

/* ==================================================================
 * == fGetInichisq : Get initial chi^2                             ==
 * == needs : Gather has to run before (0 value otherwise)         ==
 * ================================================================== 
 */
double ALIGN5_class::fGetInichisq() {
  if (gather != 0) return(sum0);
  return(0.0);
}

/* ==================================================================
 * == fGetFinchisq : Get initial chi^2                             ==
 * == needs : Solve has to run before (0 value otherwise)          ==
 * ================================================================== 
 */
double ALIGN5_class::fGetFinchisq() {
  if (solve != 0) return(fichi2);
  return(0.0);
}


/* ==================================================================
 * == fGetParmap : Get text map of parameter list                  ==
 * == needs : Prepare has to run before (0 value otherwise)        ==
 * ================================================================== 
 */
int ALIGN5_class::fGetParmap() {
  int i,active=1,op,offset;
  char textout[15][8];

  if (prepare != 1) return(0); // prepare before

  printf("\nVDS-Align: Class_ALIGN5: Parameter map:\n");

  int deza = 0;int dezb = 0;
  for (int isl=0;isl<8;++isl) {
    if (isl<3) {
      deza = 1;dezb=isl;
    } else {
      deza = isl-1;dezb=0;
    }
    for (int isec=0;isec<4;++isec) {
      // set everything to dead-state:
      for (i=0;i<15;++i) sprintf(textout[i],"       ");

      // fill the data:
      op = (1000*deza)+(30*isec)+dezb; // the sl/sect number
      for (int c=0;c<npars;++c) {// loop over all parameters
	if (((gede[c]->side)/10) == op) { 
	  if (((gede[c]->side)%10) < 2) offset=0;
	                           else offset=4;
	  if (type[c] == 2) sprintf(textout[offset + 2],"R= %3d",c); 
	  if (type[c] == 3) sprintf(textout[offset + 3],"L= %3d",c); 
	  if (type[c] == 1)
	    if ((gede[c]->lay == LAY_N1) || (gede[c]->lay == LAY_N2)) 
	      sprintf(textout[offset],"Un= %3d",c);
	    else 
	      sprintf(textout[offset+1],"Un= %3d",c);
	}
      }
      
      // check for fixed stuff:
      for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it)
	if ((((it->second->side)/10) == op) && (it->second->status == -1)) {
	  if (((it->second->side)%10) < 2) offset=0;
	                              else offset=4;
	  for (int f=offset;f<offset+4;++f)
	    sprintf(textout[f],"Fix    ");
	}

      // finally print out:
      if (active==1) {
	printf("%d :  %7s %7s %7s %7s   %7s %7s %7s %7s\n",op,
	       textout[0],textout[1],textout[2],textout[3], 
	       textout[4],textout[5],textout[6],textout[7] );
      }
    } // loop over sectors
  } // loop over all possible sides ever
  return(1);
}


/* ==================================================================
 * == fPrepare : Prepare structures for align5                     ==
 * == Parameter: Pointer to HOLMES structure                       ==
 * ==            Lateral-flag (0=Do not align lateral)             ==
 * ==            Axial-flag (0=Do not align in z)                  ==
 * ==            Rotational-flag (0=Do not align stereo angle)     ==
 * ==            Array with plane ID's wich stay fixed             ==
 * ==                                                              ==
 * == needs : ----                                                 ==
 * ==                                                              ==
 * ================================================================== 
 */

int ALIGN5_class::fPrepare(HOLMES_class *inHOLMESp, 
			   int lat, int axi, int rot, 
			   int noalign[],
			   int (*store)(   int, int, int, double  ),
			   int (*retrieve)(int, int, int, double *) ) {
  int i,j,FirstPrepare;
  int maxgede = 0;
  Gede_t *hgedeP;

  printf("VDS-Align: Class_ALIGN5: Preparing Alignment..\n");

  // call destructor already prepared
  if (prepare > 0) {
    //printf("Call fClear\n");
    fClear();
  }
  //printf("1\n");
  if (prepare ==0) FirstPrepare =1;
             else  FirstPrepare =0;
  prepare=0;
  gather =0;
  solve  =0;
  npars  =0;
  sum0   =0.0;  
  dgf    =0; 

  // ensure that HOLMES is ok:
  if (inHOLMESp == NULL) {
    fError_out("HOLMES pointer is invalid!");
    return(0);
  }
  HOLMESp=inHOLMESp;

  SEEDp   = HOLMESp->seedP();  
  if (SEEDp == NULL) {
    fError_out("SEED pointer is invalid!");
    return(0);
  }

  GEDEp   = HOLMESp->gedeP();  
  if (GEDEp == NULL) {
    fError_out("GEDE pointer is invalid!");
    return(0);
  }
  // fine all info resurces are availible

  // book the store/retriev Pointers (external access)
  Pstore    = store;
  Pretrieve = retrieve;

  // treat the noalign-stuff:
  // flag those planes which are held fixed in the alignment
  for(int i=0; i<MAXGEDE; ++i) {
    if (noalign[i] == 0) break;
    if( GEDEp->gedeSm.find(noalign[i]) == GEDEp->gedeSm.end() ) { 
      fError_out("Attempt to hold fixed a non existing plane plane!");
      ARTE_exit( GRR_SFATAL,
		 "VDS-Align: Class_ALIGN5:Attempt to hold fixed a non existing plane plane!", 0 );
    }
    if( (GEDEp->gedeSm[noalign[i]]->status) != 0) { 
      hgedeP = GEDEp->gedeSm[ noalign[i] ];
      hgedeP->status = -1;
      if ((hgedeP->pair) > 0 )
	if( (GEDEp->gedeSm.find(hgedeP->pair) != GEDEp->gedeSm.end() ))
	  GEDEp->gedeSm[hgedeP->pair]->status = -1;
    } else {
      fError_out("Attempt to hold fixed a dead plane!");
      ARTE_exit( GRR_SFATAL, "VDS-Align: Class_ALIGN5: Attempt to hold fixed a dead plane!", 0 );
    }
  }
   
  // do only for initial:
  if (FirstPrepare == 1) {
    for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
      hgedeP = it->second;
      gedeadd[it->second->index].status = it->second->status;
      gedeadd[it->second->index].paru = 0.0;
      gedeadd[it->second->index].pars = 0.0;
      gedeadd[it->second->index].parz = 0.0;
    } // of gede loop 
  } // of if (FirstPrepare == 1)
  

  // calculate the number of gede planes:
  for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it)
    maxgede++;  

  for (i=0;i<MAXGEDE;++i) { // initalise gedeadd for pos-sPointer
    gedeadd[i].matPu=-1;
    gedeadd[i].matPs=-1;
    gedeadd[i].matPz=-1;
  }
    
  // set fit-type to default:
  dofit = LATERAL*lat + AXIAL*axi + ROTATIONAL*rot;

  // allocate space for par vector :
  pars = (double **)malloc(3 * maxgede * sizeof(double *));
  gede = (Gede_t **)malloc(3 * maxgede * sizeof(Gede_t *));
  type = (int *)malloc(3 * maxgede * sizeof(int *));

  int np = 0;

  // new parameter determination from gede:
  //    MBr, 16.11.99
  int deza = 0;int dezb = 0;int dsflag=0;
  for (int isl=0;isl<8;++isl) {
    if (isl<3) {
      deza = 1;dezb=isl;
    } else {
      deza = isl-1;dezb=0;
    }
    for (int isec=0;isec<4;++isec) {
      for (int ilay =0;ilay<4;++ilay) {
	// make a side ID (10x pots!)
	int ids=(10000*deza)+(300*isec)+(10*dezb)+ilay;
	
	if (GEDEp->gedeSm.find(ids) != GEDEp->gedeSm.end()) {
	  hgedeP=GEDEp->gedeSm[ids];
	  //	  printf("id=%d\n",hgedeP->side);
	  if (dsflag==1) {
	    if ((ilay==1) || (ilay==3)) { // second sides in z only
	      if (hgedeP->status > 0) {
		// store lateral parameter only	      
		if ((dofit) & LATERAL) {
		  //linpars:
		  pars[np] = &(gedeadd[hgedeP->index].paru);
		  gede[np] = hgedeP;
		  type[np] = 1;
		  gedeadd[hgedeP->index].matPu = np;
		  np++; // number of pars increased
		} // of LATERAL, 2nd
	      } // of status >0
	    } // of second sides only
	  } else {
	    if (hgedeP->status > 0) {
	      if ((ilay==0) || (ilay==2)) dsflag = 1;
	      // store lateral, rotational and axial:
	      if ((dofit) & LATERAL) {
		//linpars:
		pars[np] = &(gedeadd[hgedeP->index].paru);
		gede[np] = hgedeP;
		type[np] = 1;
		gedeadd[hgedeP->index].matPu = np;
		np++; // number of pars increased
	      } // of LATERAL
	      // rotpars:
	      if ((dofit) & ROTATIONAL) {
		pars[np] = &(gedeadd[hgedeP->index].pars);
		gede[np] = hgedeP;
		type[np] = 2;
		gedeadd[hgedeP->index].matPs = np; 
		if (hgedeP->pair != 0)
		  if ((GEDEp->gedeSm[hgedeP->pair]->status) ==1) 
		    gedeadd[GEDEp->gedeSm[hgedeP->pair]->index].matPs = np;       
		np++; // increase number of vars
	      } // of ROTATIONAL
	      
	      // longitudinal pars:
	      if ((dofit) & AXIAL) {
		pars[np] = &(gedeadd[hgedeP->index].parz);
		gede[np] = hgedeP;
		type[np] = 3;
		gedeadd[hgedeP->index].matPz = np; 
		if (hgedeP->pair != 0)
		  if ((GEDEp->gedeSm[hgedeP->pair]->status) ==1)
		    gedeadd[GEDEp->gedeSm[hgedeP->pair]->index].matPz = np;
		np++; // increase number of vars    
	      } // of AXIAL
	    } // of status > 0
	  } // of dsflag != 1 => first side of DS	   
	} // only existing sides
	if ((ilay==1) || (ilay==3)) dsflag=0;// after layer 1==p1 or layer3 ==n2
      } // loop over layers
    } // loop over sectors
  } // loop over all possible sides ever

  printf("VDS-Align: Class_ALIGN5: Align-prepare found %3d parameters to align.\n",np);

  // save npars/np :
  npars = np;
  // allocate and initialise the matrix:
  amat = (double *)malloc(np * np * sizeof(double));
  atam = (double *)malloc(np * np * sizeof(double));
  mp = (int *)malloc(np * sizeof(int));
  // allocate and initialise the sum and solution vector:
  avec   = (double *)malloc(np * sizeof(double));
  svec   = (double *)malloc(np * sizeof(double));
  esvec  = (double *)malloc(np * sizeof(double));
  //  svvec  = (double *)malloc(np * sizeof(double));
  avsing = (double *)malloc(np * sizeof(double));

  // allocate and initialise the vector:
  hiscount = (int *)malloc(np * sizeof(int));
    //  ...yes the program looks big from here... :-))

  // set all entries to zero: 
  int matposi = 0;
  for (i=0;i < (np);++i) {
    avec[i]      = double(0.0);
    mp[i]        = np*i;
    (*(pars[i])) = double(0.0);
    hiscount[i]  = 0;
    for (j=0;j < (np);++j){
      amat[matposi] = double(0.0);
      atam[matposi] = double(0.0);
      ++matposi;
    }
  }

  prepare=1;  // now we are initialised!
  return(1);
}


/* ==================================================================
 * == fGather : Accumulate the infos from tracks                   ==
 * == needs   : fPrepare has to be run before                      ==
 * ==           int tr : track no in HOLMES current list           ==
 * ==                                                              ==
 * ================================================================== 
 */
int ALIGN5_class::fGather(int tr) {
  Gede_t       *hgedeP;           // Pointers to gede structures
  char         seed2hit[MAXGEDE]; //gedeindex as index -> =1 if hit in plane
  float        seedu[MAXGEDE];    // "         "   "      =u for that     "
  float        tpn[5];            // track-projektor residual strips (rot,z)

  Seed_t       *seedP;            // pointer to seeds
  Hitb_t       *hit;              // Pointer to hitb structy of holmes 
  int          i,j;  // std counter
  int          maxhit,chit;       // hits on track

  double       da[15];            // the projectors lateral,rotational,axial
  int          no[15];            // number of related matrix entry
  double       vu0;               // the chi2 contrib of act ReLe
  Track_t      track;             // track, build out of seeds during aligng
  Track_t      strack;            // track, build out of seeds during aligng

  int          sindex[NSEED];     // index to the four seeds


  //printf("prep=%d   ",prepare);

  if (prepare != 1) return(0); // prepare before

  track=*(HOLMESp->trackP(tr));
  maxhit = track.nhit;
  
  //  printf("ntr=%d  %d\n",tr,maxhit);

  // just avoifd the simplest case of bad data (R..)
  if (track.nhit < NSEED+1) return(0);

  // save number of measurements :
  dgf += maxhit-4;  
    
  // clear old hitmap :
  for (i=0;i < MAXGEDE;seed2hit[i++]=0);
  // build map for hits :
  for (i = 0; i<maxhit; ++i ) {
    // write 1 if index is used in track	
    seed2hit[HOLMESp->hitbP(track.hit[i])->gede->index] = 1;	
    // get the u's out :	     
    seedu[HOLMESp->hitbP(track.hit[i])->gede->index] = 
      HOLMESp->hitbP(track.hit[i])->u;
  }

  //printf("t g1\n");

  // loop over all possible seedings
  for(Seed_mi it=SEEDp->seedM.begin();it != SEEDp->seedM.end();++it) {
    seedP=it->second;
    for (i=0;i<NSEED;++i) sindex[i] = (seedP->gedeP)[i]->index;
    // accept seeding
    if (seed2hit[sindex[0]] == 0) continue;
    if (seed2hit[sindex[1]] == 0) continue;
    if (seed2hit[sindex[2]] == 0) continue;
    if (seed2hit[sindex[3]] == 0) continue; // seeding ok!	

    //set up pars for seed combi
    strack.x = strack.y = strack.tx = strack.ty = 0.0;
    for (i=0; i<4; ++i) {
      float u = seedu[sindex[i]];
      strack.x += (seedP->vx )[i] * u;
      strack.y += (seedP->vy )[i] * u;
      strack.tx+= (seedP->vtx)[i] * u;
      strack.ty+= (seedP->vty)[i] * u;
    }

    // NO confirmation is done (conf should have been done by holmes
    // loop over remainig hits ( #>0 )
    for(chit=0; chit<maxhit; ++chit) {
      hit = HOLMESp->hitbP(track.hit[chit]); // get hit -------------------
      // hit used for seed : (comment for zero enriched sample!)
      if ((hit->gede->index == sindex[0]) ) continue;
      if ((hit->gede->index == sindex[1]) ) continue;
      if ((hit->gede->index == sindex[2]) ) continue;
      if ((hit->gede->index == sindex[3]) ) continue;
      
      // projector from strack to act. resplane:
      // Calc : u = (p==tpn)_vec * t_vec
      tpn[0] = hit->gede->v[0];//   hit->gede->c ;
      tpn[1] = hit->gede->v[1];// -(hit->gede->s);
      tpn[2] = hit->gede->v[2];// tpn[0] * hit->gede->z;
      tpn[3] = hit->gede->v[3];// tpn[1] * hit->gede->z;
      
      for (i=0; i<4; ++i) {  // work on all four seedplanes
	hgedeP = (seedP->gedeP)[i]; // points to gede of act. seed plane
	no[i   ]  = gedeadd[hgedeP->index].matPu;
	no[i+ 5]  = gedeadd[hgedeP->index].matPs; 	    
	no[i+10]  = gedeadd[hgedeP->index].matPz; 	    
	/* if (no[i]    == -1) printf("SE -1: i=%d, side=%d, \n",
				i,hgedeP->side,hgedeP->index,gedeadd[hgedeP->index].matPu);
	   if (no[i+5]  == -1) printf("SE -1: i=%d, side=%d, \n",
				i,hgedeP->side,hgedeP->index,gedeadd[hgedeP->index].matPs);
	   if (no[i+10] == -1) printf("SE -1: i=%d, side=%d, \n",
				i,hgedeP->side,hgedeP->index,gedeadd[hgedeP->index].matPz);
	*/

	// set up v0 (from seeding to act srips) = tp0*Vseed 
	// calc v_vec = p_vec * M_mat^-1
	da[i]     = double(
			   tpn[0] * (seedP->vx )[i] +
			   tpn[1] * (seedP->vy )[i] +
			   tpn[2] * (seedP->vtx)[i] +
			   tpn[3] * (seedP->vty)[i]  
			   );
	// set up s : (rotation)
	da[i+ 5]  = double(
			   (strack.x  * (hgedeP->s) )
			   +(strack.y  * (hgedeP->c) )
			   +(strack.tx * (hgedeP->s) * (hgedeP->z))
			   +(strack.ty * (hgedeP->c) * (hgedeP->z))
			   );
	//da[i+ 5] *= double(da[i]);
	da[i+ 5] *= double(da[i]*RSFACT);
	// set up z : (z-shift) == -v[i] * t * r
	da[i+10]  = double((
			    -da[i] * ( (strack.tx * (hgedeP->c) )
				       -(strack.ty * (hgedeP->s) ) )
			    )*ZSFACT );
      }  // of seedplane loop
      
      no[ 4] = gedeadd[hit->gede->index].matPu;
      no[ 9] = gedeadd[hit->gede->index].matPs;
      no[14] = gedeadd[hit->gede->index].matPz;
      /*if (no[ 4] == -1) printf("SE -1: i=%d, side=%d, \n",
			       i,hgedeP->side,hgedeP->index,gedeadd[hgedeP->index].matPu);
	if (no[ 9] == -1) printf("SE -1: i=%d, side=%d, \n",
			       i,hgedeP->side,hgedeP->index,gedeadd[hgedeP->index].matPs);
        if (no[10] == -1) printf("SE -1: i=%d, side=%d, \n",
			       i,hgedeP->side,hgedeP->index,gedeadd[hgedeP->index].matPz);
      */
      
      //accumulate actual residual :
      vu0    = tpn[0] * strack.x  + tpn[1] * strack.y  + 
	tpn[2] * strack.tx + tpn[3] * strack.ty  ;
      vu0   -= (hit->u);  // 5. element explicit -1 == residual
      // set up 5. element for local stuff on res plane
      // lateral 5. element v ==-1 :
      da[4]  = double(-1.0);                           
      // 5. element s == -q_vec * t_vec (rotation) :
      da[9]  = double((
		       -(strack.x  * (hit->gede->s) ) 
		       -(strack.y  * (hit->gede->c) )
		       -(strack.tx * (hit->gede->s) * (hit->gede->z))
		       -(strack.ty * (hit->gede->c) * (hit->gede->z))
		       )*RSFACT);
      // axial, 5. element == -sum zi :
      da[14] = double(ZSFACT*(-da[10]-da[11]-da[12]-da[13]));                   
      	  
      // accumulate to table/vetcor : -----------------------------------
      int    ni;
      double we,dai;

      //printf("t1\n");

      if (Pstore == NULL) { // ------------- INTERNAL!! --------------------
  //printf("ti\n");

	for (i=0;i<15;++i) { // loop over da (matrix is bigger!)
	  ni=no[i];                     // line of mat/vec to put in ith proj.
	  if (ni == -1) continue;       // only valid planeparas!
	  dai=da[i];
	  for (j=i+1;j<15;++j) {          // off diagonal
	    if ( no[j] == -1) continue; // only valid planeparas!
	    we = dai*da[j];
	    amat[mp[   ni] + no[j]] += we;
	    amat[mp[no[j]] +    ni] += we;
	  }
	  amat[mp[ni] + ni] += dai*dai; // diagonal
	  avec[ni]                  += dai*vu0; // vector
	  hiscount[ni] +=1; 
	}
	++gedeadd[hit->gede->index].rc; // rele for that plane found 
	++nreles;
	sum0 += vu0*vu0; // sum up for initial Alignment
      } else {              // ------------ EXTERNAL (online case) ----------
  //printf("t 22\n");
	for (i=0;i<15;++i) { // loop over da (matrix is bigger!)
	  ni=no[i];                     // line of mat/vec to put in ith proj.
	  if (ni == -1) continue;       // only valid planeparas!
	  dai=da[i];
	  for (j=i+1;j<15;++j) {          // off diagonal
	    if ( no[j] == -1) continue; // only valid planeparas!
	    we = dai*da[j];
	    //amat[mp[   ni] + no[j]] += we;
	    Pstore(ALIGN5_SMAT,ni,no[j],we);
	    //amat[mp[no[j]] +    ni] += we;
	    Pstore(ALIGN5_SMAT,no[j],ni,we);
	  }
	  //amat[mp[ni] + ni] += dai*dai; // diagonal
	  Pstore(ALIGN5_SMAT,ni,ni,dai*dai);
		
	  Pstore(ALIGN5_SVEC,ni,0,dai*vu0); // vector
	  hiscount[ni] +=1; 
	}
	// Pstore(1,i,j,du) -> Matrix, store(2,i,0,du) -> Vector,  
	++gedeadd[hit->gede->index].rc; // rele for that plane found
	Pstore(ALIGN5_SVEC,ALIGN5_RPO,0,double(1.));
	++nreles;
	Pstore(ALIGN5_SVEC,ALIGN5_CPO,0,vu0*vu0);
	sum0 += vu0*vu0; // sum up for initial Alignment
      }
    } // res hits on track    
  } // seeding loop      
  //printf("t\n");

  if (Pstore == NULL) {
    ntrack ++;
    gather ++;
  } else 
    Pstore(ALIGN5_SVEC,ALIGN5_TPO,0,double(1));
  //printf("prep=%d\n",prepare);
  return(1);
}

/* ==================================================================
 * == fSolve : Invert the algnment matrix, calc the pars           ==
 * == needs : fGather has to be run before at least 10 times       ==
 * ==                                                              ==
 * == Using SVD to get the result!                                 ==
 * ================================================================== 
 */
int ALIGN5_class::fSolve() {
  int i,mpi; 
  int info;
  double stored;

  if (Pretrieve != NULL) {
    //printf("AVIATION: external storage!\n");
    //printf("ntrack=%d, nreles=%d, sum0 = %8.2f\n",ntrack,nreles,sum0);
    for (int i=0;i<npars;++i) {
      for (int j=0;j<npars;++j) {
	Pretrieve(ALIGN5_SMAT,i,j,&stored);
	  amat[mp[i] + j] = stored;
      }
      Pretrieve(ALIGN5_SVEC,i,0,&stored);
      avec[i] = stored;
    }
    Pretrieve(ALIGN5_SVEC,ALIGN5_RPO,0,&stored);
    nreles = int(stored);

    Pretrieve(ALIGN5_SVEC,ALIGN5_CPO,0,&stored);
    sum0 = float(stored);

    Pretrieve(ALIGN5_SVEC,ALIGN5_TPO,0,&stored);
    ntrack = int(stored);

    gather = ntrack;
    //printf("%d %d %d   ntrack=%d, nreles=%d, sum0 = %8.2f\n",
    //	   ALIGN5_RPO,ALIGN5_CPO,ALIGN5_TPO,ntrack,nreles,sum0);
  }

  if (gather < MINGATHER) return(0); // accumulate before
  //printf("g prep=%d\n",prepare);

  printf("Solve Alignment Matrix :\n");

  double *svmat = (double *)malloc(npars * npars * sizeof(double));
  // save initial matrix :
  mpi = 0;
  for (i=0;i<npars*npars;++i) {
   svmat[mpi] = amat[mpi];
   atam[mpi] = amat[mpi++];
  }

  //printf("Searching for eigenvalues of matrix..\n");
  // do the Singular Value Decomposition:
  //eigenvalues(svmat,svvec,np);
  //printf("done !\n");
  

  // number of degrees of freedom = measurements (tracks*hits-seed(=4) - pars
  dgf -=npars;
  // printf("initial chisq : %8.4f, #dgf = %6d\n",sum0,dgf);  
  printf("Inverting matrix..");fflush(stdout);

  // copy b0 to solution (solvesvd will take b0 from it and write the solution back.
  for (i=0;i<npars;++i) {
    svec[i] = -avec[i];
  }

  solvesvd(atam, svec, npars, avsing, T_SINGULAR,   &rank,  &info );
  if (info != 0) {
    printf("\nVDS-Align: Class_ALIGN5: Matrix inversion using SVD failed, call seems to be corrupt !!\n\n");
    printf(" Return values : info=%d\n ",info);
    printf("Pre mortem parameters :\n");
    for (i=0;i<npars;++i) {
      printf("Parameter %2d, %d, type = %d : ",i,
	     gede[i]->side,type[i]);
      printf("mii = %12.4f, ",amat[mp[i] + i]);
      //printf("EiV = %14.6f ",svvec[i]);
      printf("#rh = %8d\n",hiscount[i]);
    }
    printf("Exiting now.\n");
    ARTE_exit( GRR_SFATAL,
	       "VDS-Align: Class_ALIGN5: Matrix inversion using SVD failed", 0 );
  }
  printf("done !\n");

	 
  // calculate the solution-vector:
  for (i=0;i<npars;++i) {
    (*(pars[i])) = svec[i]; // copy pars to gedeadd struct
  }
  solve=1;
  // from here the data is known!!
  fichi2 = fCalcchisq();

  return(1);
}

/* ==================================================================
 * == fCalcchisq : Calculate a chi^2 from the data for new pars    ==
 * == needs : gather has to be run before                          ==
 * == return: The actual chi^2 value                               ==
 * ==                                                              ==
 * ================================================================== 
 */
double ALIGN5_class::fCalcchisq() {
  int i,j,smp;
  double chisq,w;
  if (gather < 1) return(double(0)); // accumulate before

  chisq = sum0;
  for (i=0;i<npars;++i) {
    chisq += 2 * (*(pars[i])) * (avec[i]);
    smp =mp[i]; // semiPointer to row i of amat
    w=0.0;
    for (j=0;j<npars;++j)
      w += amat[smp + j] * (*(pars[j])); 
    chisq += w * (*(pars[i]));
  }
  return(chisq);
} // =============== END of calcchisq member func. =====================



/* ==================================================================
 * == fOut   : Output of the alignment parameters                  ==
 * == needs : fSolve has to be run before                          ==
 * ==                                                              ==
 * ==                                                              ==
 * ================================================================== 
 */
int ALIGN5_class::fOut() {
  int i;

  if (solve != 1) return(0); // solve before
  
  for (i=0;i<npars;++i) {
    // calculate the errors :
    double k1=(avec[i]) - (amat[mp[i] + i])*(sum0-fichi2-1);
    double k2 = 0.0;
    double k3 = 0.0;
    if (k1 >0) {
      k2= (-avec[i] + sqrt(k1))/amat[mp[i] + i];
      k3= (-avec[i] - sqrt(k1))/amat[mp[i] + i];
    } 
    k2 = (k2 - k3)/2*sqrt(fichi2/dgf); // take the mean value for error, scale it to 
                                     // chi2/ndgf == 1
    
    // output results:
    printf("%4d %d ",i,gede[i]->side);
    switch (type[i]) {
    case 1 : printf("LATERAL    p = %10.6f,   (%9.2f ",svec[i],
		    (svec[i]     ) * LENSC);
             if (k1>0) {
	       printf("+- %7.2f)my    ",k2 * LENSC);
	     } else {
	       printf("-+ ******.**)my    ");
	     }
             break;
    case 2 : printf("ROTATIONAL p = %10.6f,   (%9.2f ",svec[i]*RSFACT,
		    (svec[i]) * 180000/Pi*RSFACT);
             if (k1>0) {
	       printf("+- %7.2f)mDeg  ",
		      k2 * 180000/Pi*RSFACT );
	     } else {
	       printf("+- ******.**)mDeg  ");
	     }
             break;
    case 3 : printf("AXIAL      p = %10.6f,   (%9.2f ",svec[i],
		    (svec[i]     ) * LENSC*ZSFACT);
             if (k1>0) {
	       printf("+- %7.2f)my    ",
		      k2 * LENSC*ZSFACT);
	     } else {
	       printf("+- ******.**)my    ");
	     }
             break;
    }
    printf("mii = %12.4f, ",amat[mp[i] + i]);
    printf("SV = %11.4f, %7.2E",avsing[i], avsing[i]/avsing[0]);
    printf("\n");
  }
  printf("Singular Value decomposition : rank=%2d\n",rank);

  return(1);
} // =============== END of fOut member func. ========================================




/* ==================================================================
 * == fChange_GEDE : Change the entrire GEDE structure with        ==
 * ==                alignment corrections                         ==
 * == needs : fSolve has to be run before                          ==
 * ==                                                              ==
 * ================================================================== 
 */
int ALIGN5_class::fChange_GEDE() {
  Gede_t *hgedeP,*gede2P = NULL;
  
  if (solve != 1) return(0); // solve before

  if (GEDEp == NULL) return(0);

  // first go over parameters to copy rot/z-changes to partners
  // NO precaution for dead status=0 planes is done!
  for (int np=0;np<npars;++np) {
    hgedeP=gede[np];
    if ((hgedeP->pair !=  0) && (GEDEp->gedeSm.find(hgedeP->pair) != GEDEp->gedeSm.end() )) {
     if (type[np] == 2) gedeadd[GEDEp->gedeSm[hgedeP->pair]->index].pars  = gedeadd[hgedeP->index].pars;
     if (type[np] == 3) gedeadd[GEDEp->gedeSm[hgedeP->pair]->index].parz  = gedeadd[hgedeP->index].parz;
    }
  }

  for (GedeI_mi it = GEDEp->gedeSm.begin();it != GEDEp->gedeSm.end(); ++it) {
    hgedeP = it->second;
    // gede changes :          
    hgedeP->z   += float(gedeadd[hgedeP->index].parz*ZSFACT);
    hgedeP->da  += float(180/Pi*(gedeadd[hgedeP->index].pars)*RSFACT);

    hgedeP->sxmn += float( (hgedeP->c) * (gedeadd[hgedeP->index].paru));  
    hgedeP->sxmx += float( (hgedeP->c) * (gedeadd[hgedeP->index].paru)); 
    hgedeP->symn += float(-(hgedeP->s) * (gedeadd[hgedeP->index].paru));  
    hgedeP->symx += float(-(hgedeP->s) * (gedeadd[hgedeP->index].paru));
    if ((hgedeP->pair !=  0) && (GEDEp->gedeSm.find(hgedeP->pair) != GEDEp->gedeSm.end() )) {  
      gede2P = GEDEp->gedeSm[hgedeP->pair];
      hgedeP->sxmn += float( (gede2P->c) * (gedeadd[gede2P->index].paru));  
      hgedeP->sxmx += float( (gede2P->c) * (gedeadd[gede2P->index].paru)); 
      hgedeP->symn += float(-(gede2P->s) * (gedeadd[gede2P->index].paru));  
      hgedeP->symx += float(-(gede2P->s) * (gedeadd[gede2P->index].paru));
    }   
    
    // the other coordinates need an update:

    // set the dmin (independent from rotations!) :
    float c1 = cos( hgedeP->alpha * Pi / 180. );
    float s1 = sin( hgedeP->alpha * Pi / 180. );
    hgedeP->dmin  = hgedeP->d0 + hgedeP->sxmn * c1 - hgedeP->symn * s1;
    
    // cos,sin of 'real' stereo angle
    hgedeP->c = cos( ( hgedeP->alpha + hgedeP->da ) * Pi / 180. ); 
    hgedeP->s = sin( ( hgedeP->alpha + hgedeP->da ) * Pi / 180. ); 
    
    hgedeP->v[0]  =  hgedeP->c;
    hgedeP->v[1]  = -hgedeP->s;
    hgedeP->v[2]  =  hgedeP->c * hgedeP->z;
    hgedeP->v[3]  = -hgedeP->s * hgedeP->z;
    for (int i=0; i<4; i++)
      for (int j=0; j<4; j++)
	hgedeP->vv[4*i+j] = hgedeP->v[i] * hgedeP->v[j];
    
  } // 2. loop (over GEDE)
  return(1);
} // =============== END of fChange_GEDE member func. ===============================


void ALIGN5_class::fError_out(const char *s) {
  /*====================================
     Give output to the various sources
   =====================================
   */

  fprintf(stderr,"VDS-Align: Class_ALIGN5: Fatal Error : %s!!\n",s);
  
  return;
} // =============== END of fOut member func. ========================================


//------------------------------------------------ Low level matrix:

int ALIGN5_class::sqrinvd(double *a, int n) {
  /*====================================
   invert a square matrix
   ====================================
   */

  int   ifail;
  double *work = (double*)malloc(n*sizeof(double));
  dinv_(&n, a, &n, work, &ifail);
  free(work);
  return(ifail);
} 


int ALIGN5_class::solveortho(double *matrix, double *solution, double *result, int n) {
  // NOT IN USE !!!
  // complete orthogonel factorization solution
  //
  //dgelsx_(int *, int *, int *, double *, int *, double *, int *, int *, 
  //	  double *, int *, double *, int *);
 return(0);
}


// Solve the matrix * x == solution equation
// matrix   : double array n x n
// solution : The right-hand vector on input, x-vector on output 

int ALIGN5_class::solvesvd(double *matrix,     // input matirx n x n
	     double *solution,   // solution vector b and result x after
	     int    n,           // dimension of m and b
	     double *singval,    // after: sing. values
	     double rcond,       // number below wich a singular is assumed
	     int    *rank,       // resulting effective rank of a
	     int    *info    ) { // =0 all ok, <0 i.th ele illegal, 
                                 // >0 i off-diag non converged

  // Compute the minimum norm solution for a least squares problem.
  // Method : Singular value decomposition

  int nrhs  = 1;     // solve only for one vector                
  int lda   = n;     // leading (used) dimension of matrix
  int ldb   = n;     // leading dimansion of b solution
  int lwork = 10*n;  // length work-space
  double work[10*n]; // workspace

  dgelss_(&n, &n, &nrhs, matrix, &lda, solution, &ldb, singval,
	  &rcond, rank, work, &lwork, info);

  if (*info !=0 ) printf("\nsolvesvd: lwork=%d,info =%d\n",lwork,*info);
  return(0);
}


int ALIGN5_class::eigenvalues(double *matrix, double *result, int n)
{ 
  int   lda,lwork,info;
  char  uplo[2] = "U";
  double e[n];    // 1. off diagonal in result
  double tau[n];  // 
  double work[20*n]; //

  lda = n;
  lwork = n;
  // 1. the tridiagonalisation step: 
  // Method : orthogonal similarity  transformation
  dsytrd_(uplo, &n, matrix, &lda, result, e, tau, work, &lwork, &info);
  if (info < 0) printf("Align-Eigenvalues : Error of parameter %d!\n",info);
  printf("VDS-Class_ALIGN5: dsytrd ready, info = %d, opt val. : %4.0f\n",info,work[0]);

  // 2. The solve step: 
  // Method : Pal-Walker-Kahan modified QL / QR 
  dsterf_(&n, result, e, &info);
  printf("VDS-Class_ALIGN5: dsterf ready, info = %d\n",info);
  return(1); 
}



#endif // end of code definiton

// =========== EOF ===== EOF ===== EOF ===== EOF ===== EOF ===== EOF =============




