

#include "care/ECDcalIO.hh"
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include "care/ELNearCells.hh"
#include "care/ECDbNoise.hh"
#include "etb.h"
#include <algorithm>
#include <math.h>

// extern "C" void unsetenv(const char *name);
extern "C" void putenv(const char *name);


class ECSmplStat {
protected:
    int	n;
    double x;
    double x2;
    double minValue, maxValue;

    public :

    ECSmplStat();
    ECSmplStat(const ECSmplStat & stat);
    ~ECSmplStat();
    const ECSmplStat & operator=(const ECSmplStat & stat);
    void reset(); 

    void operator+=(double);
    int  samples() const;
    double mean() const;
    double stdDev() const;
    double var() const;
    double min() const;
    double max() const;
};

inline ECSmplStat::ECSmplStat(){ reset();}
inline int ECSmplStat::samples() const {return(n);}
inline double ECSmplStat::min() const {return(minValue);}
inline double ECSmplStat::max() const {return(maxValue);}

ECSmplStat::ECSmplStat(const ECSmplStat & stat){
  operator=(stat);
}

ECSmplStat::~ECSmplStat(){}

const ECSmplStat & 
ECSmplStat::operator=(const ECSmplStat & stat){
  n = stat.n;
  x = stat.x;
  x2 = stat.x2;
  minValue = stat.minValue;
  maxValue = stat.maxValue;
  return *this;
}

/// reset internal data
void ECSmplStat::reset(){
  n = 0;
  x = 0.0;
  x2 = 0.0;
  minValue = HUGE_VAL; // +infinity
  maxValue = -HUGE_VAL; // -infinity
}

// add operator
void ECSmplStat::operator+=(double val){
  n++;
  x += val;
  x2 += val*val;
  if( val>maxValue ) maxValue = val;
  if( val<minValue ) minValue = val;
}

/// return the mean of the sample.
double ECSmplStat::mean() const {
  return (n>0) ? x/(double)(n) : 0.0;
}

/// return the standard deviation of the sample.
double ECSmplStat::stdDev() const {
  return sqrt( var() );
}

/// return the standard deviation of the sample.
double ECSmplStat::var() const {
  double mn = mean();
  double vr = (n>0) ? x2/n-mn*mn : 0.0 ;
  if( vr<0.0 ) vr = 0.0;
  return vr;
}


int ECCorPed(EcalEvent_t * start){

  enum {minfed=16,maxfed=241,ncells=5956,maxadc=500};
  static ECDbNoise* dbnoise = 0;
 
  if( dbnoise==0 ){
    // read the DB noise
    char * dbn = getenv("HBECALNOISE");
    if( dbn==0 ) dbn = "/ECAL_CALIB";
    dbnoise = new ECDbNoise();
    dbnoise->openDB(dbn,"ECALNOISE");
    int rc = 0;
    int major=0,minor=0;
    int run =ArtePointer<EVHD>(1)->run;
    if( run>=14988 && run<15707 ){
      major = 16206;
      minor = 0;
    } else if( run>=15707 && run<17268 ){ 
      major = 15100;
      minor = 0;
    }
    if( major==0 ){
      putenv("ECALNOISE=no");
//      unsetenv("ECALNOISE");
      return 1;
    }
    rc = dbnoise->readFromDB(major,minor,TABLEeq);
    if( rc!=226 ){
      arteMsg<<" Error opening ECAL DB noise "<<dbn<<" ("
	     <<major<<","<<minor<<")";
      arteMsg.error();
      delete dbnoise;
      dbnoise = 0;
      putenv("ECALNOISE=no");
//      unsetenv("ECALNOISE");
      return 1;
    }
    arteMsg<<"Applying noise correction ";
    arteMsg.allp();
  }
  //
  // Step 0: read the buffer 
  //

  // Count fed occurrencies
  char occ[256];
  for(int i=0; i<256; ++i) occ[i] = 0;
  vector<ECSmplStat> adcb(maxfed);

  int adcs[ncells+1];
  for(int i=0; i<=ncells; ++i) adcs[i] = -100;
  
  unsigned short sfed_id, idmk, bx, firstbx = 0x8000;
  unsigned short channel, dummy1, dummy2, siz;
  int n=0;
  ECDcalIO dcal;
  dcal.start(start);
  // read the bank header
  dcal.getBankHeader(dummy1, dummy2); 
  unshort fed_id = dcal.nextFED(); // read the Front End Detector IDentifier
  while( fed_id>0 ) {
    unsigned int nch = dcal.getFedHeader(siz, sfed_id, idmk, bx);
    if( firstbx == 0x8000) firstbx = bx;
    if( fed_id>15 && fed_id<242){
      if( occ[fed_id]==0 ){	
	int count=0;
        for(unsigned int jj=0; jj<nch; jj++){
          int ofs = 0;
          // readout channel 
          dcal.getNextShort(ofs,channel);
	  int gcce = ArtePointer<ECBO>(fed_id-minfed+1)->chan[ofs];
	  /*	  if( gcce>2400 && gcce<2420) cout<<gcce<<" "<<channel<<" "
					  <<dbnoise->chanCorr(gcce)
					  <<" ECHF:"
					  <<ArtePointer<ECHF>(gcce)->pedf
					  <<" "
					  <<ArtePointer<ECHF>(gcce)->wid
					  <<endl; */
	  if( adcs[gcce]==-100 && dbnoise->chanCorr(gcce)>0.5) {
	    adcs[gcce] = channel;
	    ArtePointer<ECHF> echf(gcce);
	    n++;
	    if( fabs(channel-echf->pedf)<=3.*echf->wid && echf->wid>0.1 ){
	      float r = (channel-echf->pedf)/echf->wid;
	      adcb[fed_id-minfed] += r;
	    }
	  }
        }
      }
      occ[fed_id]++;
    }
    fed_id = dcal.nextFED(); // go and take the next FED
  }
  //  cout<<" Step 0: "<<n<<" cells to consider "<<endl;

  //
  // step 1: kill shower candidates
  //
  n=0;
  ELNearCells nc;
  for(int gcce=1; gcce<=ncells; gcce++){
    ArtePointer<ECHF> echf(gcce);
    /*    if( gcce>2400 && gcce<2420) cout<<gcce<<" "<<channel<<" "
				    <<adcs[gcce]<<endl; */
    if( adcs[gcce]>-98 ){
      if( fabs(adcs[gcce]-echf->pedf)>=10.*echf->wid && (echf->stat&1)==1 ){
	adcs[gcce] |= 0x10000;
	nc.of(gcce, ELNearCells::square);
	for(nc.begin(); nc.more(); nc.next()){
	  if( adcs[(*nc).index()]>0 ) n++;
	  /*	  if( (*nc).index()>2400 && (*nc).index()<2450){
	    cout<<gcce<<" kill "<<(*nc).index()<<" "<<adcs[(*nc).index()] <<endl;
	  } */
	  adcs[(*nc).index()] |= 0x10000;	
	}
      }
    }
  }
  //  cout<<" Step 1: killed "<<n<<" cells as shower member candidates"<<endl;
  // 
  // step 2: evaluate mean on boards
  //
  n=0;
  for(int gcce=1; gcce<=ncells; gcce++){
    if( adcs[gcce]>-98 && adcs[gcce]<0x10000 ){
      ArtePointer<ECHF> echf(gcce);
      if( fabs(adcs[gcce]-echf->pedf)<=3.*echf->wid && echf->wid>0.1 ){
	float r = (adcs[gcce]-echf->pedf)/echf->wid;
	adcb[ArtePointer<ECHS>(gcce)->fdid-minfed] += r;
	n++;
      }
    }
  }

  //  cout<<" Step 2: "<<n<<" cells partecipating to board means "<<endl;
  //
  // step 3: correct with board data (if enough)
  //
  n=0;
  for(int i=0; i<226; i++){
    if( adcb[i].samples()>4 ){
      ArtePointer<ECBO> ecbo(i+1);
      for(int ch=0; ch<32; ch++){
	int gcce= ecbo->chan[ch];
	ArtePointer<ECHF> echf(gcce);
	if( gcce>0 && gcce<=ncells && adcs[gcce]>-98 ){
	  adcs[gcce] = 
	    ((int)((adcs[gcce] & 0x7fff)-adcb[i].mean()*echf->wid) ) | 0x8000;
	  n++;
	}
      }
    }
  }
  //  cout<<" Step 3: "<<n<<" cells corrected!"<<endl;
  // 
  // step 4: correct with correlation data
  //
  n=0;
  for(int i=0; i<226; i++){
    if( adcb[i].samples()<5 ){
      const vector<ECFedCor> *fdv = &dbnoise->fedCorrs(i+minfed);
      if( fdv->size()>0 ){
	float means = 0.0;
	float sumcor = 0.0;
	for(vector<ECFedCor>::const_iterator it=fdv->begin(); 
	    it!=fdv->end(); ++it){
	  if( fabs(it->corr)>0.75 && adcb[it->fed-minfed].samples()>4 ){
	    means += adcb[it->fed-minfed].mean()*it->corr;
	    sumcor += fabs(it->corr);
	  }
	}
	if( sumcor>0.0 ){
	  // ok can change the data!
	  means /=sumcor;
	  ArtePointer<ECBO> ecbo(i+1);
	  for(int ch=0; ch<32; ch++){
	    int gcce= ecbo->chan[ch];
	    ArtePointer<ECHF> echf(gcce);
	    if( gcce>0 && gcce<=ncells && adcs[gcce]>-98 ){
	      adcs[gcce] = 
		((int)((adcs[gcce]&0x7fff)-means*echf->wid) ) | 0x8000;
	      n++;
	    }
	  }
	}
      }
    } else{
      //      cout<<"Fed "<<i+16<<" "<< adcb[i].samples()<<endl;
    }
  }
  //  cout<<" Step 4: "<<n<<" cells corrected!"<<endl;

  // 
  // step 5: rebuild the DCAL table for future analysis
  //
  n=0;
  for(int i=0; i<256; ++i) occ[i] = 0;
  dcal.start(start);
  // read the bank header
  dcal.getBankHeader(dummy1, dummy2); 
  fed_id = dcal.nextFED(); // read the Front End Detector IDentifier
  while( fed_id>0 ) {
    unsigned int nch = dcal.getFedHeader(siz, sfed_id, idmk, bx);
    if( firstbx == 0x8000) firstbx = bx;
    if( fed_id>15 && fed_id<242){
      if( occ[fed_id]==0 ){	
        for(unsigned int jj=0; jj<nch; jj++){
          int ofs = 0;
	  int gcce = ArtePointer<ECBO>(fed_id-15)->chan[jj];
	  if( (adcs[gcce] & 0x8000) == 0x8000 && adcs[gcce]>-98){
	    unshort value = (adcs[gcce] & 0x7FFF);
	    if( n<20 ){
	      //	      cout<<"writing in "<<fed_id<<" "<<jj<<":"
	      //	  <<(hex)<<value<<(dec)<<endl;
	    } 
	    dcal.putNextShort(value);
	    n++;
	  } else {
	    dcal.getNextShort(ofs,channel);
	  }
        }
      }
      occ[fed_id]++;
    }
    fed_id = dcal.nextFED(); // go and take the next FED
  }
  //  cout<<" Step 5: "<<n<<" cells written to DCAL!"<<endl;
  return 0;
}



