//                              -*- Mode: C++ -*-
//: EMCluster.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 3/4/97
// Last Update ........
//

#ifndef EMCLUSTER_HH_
#define EMCLUSTER_HH_

#include "care/ECALtypes.hh"
#include "care/EMClusterMember.hh"
#include "care/ECClHyp.hh"


/**
class EMCluster.
  Concrete derived class for the hierarchical cluster tree used in
  pattern recognition of ECAL. This class stores a list of children in the
  tree.

*/

class EMCluster : public EMClusterMember {
public:
  /**@name Constructors and Destructors */
  //@{
  /** Constructors.
     The construction of EMCluster is done referring to some
     parent. The depth level is adjusted accordingly. */
  EMCluster(EMCluster* parent, EMClusterMember* firstChild);
  
  ///  The default constructor builds an empty list. 
  EMCluster();

  /// Destructor:   A virtual destructor is needed.
  virtual ~EMCluster();
  //@}

  /**@name Beheavioral methods */
  //@{
  /// Reset all info in the cluster.
  void  reset();

  
  /// Add a Child to the list: it makes sense only for cluster objects.
  virtual  EMClusterMember* add(EMClusterMember* child);  
  
  ///  Start an unititialize cluster with the Child passed
  void start(EMClusterMember* child);

  /**
    Move a Child from its' parent list to this cluster and put it at the
    end. */
  EMClusterMember* moveAtEnd(EMClusterMember* child);  
 
  //--
  /// Move a Child from a EMCluster to the current one.
  virtual void move(EMClusterMember* pChild);
  /**
    After operation on cluster such as reordering, merging or other
    It is possible that the depths are not correct. With this function
    the depth of the EMClusterMember and all its children are adjusted. */ 
  virtual void adjust_depths();

  /**
    Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
    that are object with definite analyzing properties. 
    EMClusterMembers MUST ignore these properties: in this way a loose coupling
    between EMClusterMembers and EAClusterAnalyzers is achieved and both
    can vary independently during program evolution.
    Examples of EAClusterAnalyzers are those that should return
    total energy of a cluster, the Center of Gravity, the Number of Cells. */
  virtual void analyzeMe(EAClusterAnalyzer & analyzer) const;

 /** 
    Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
   that are object with definite analyzing properties.
   EMClusterMembers MUST ignore these properties: in this way a loose coupling
   between EMClusterMembers and EAClusterAnalyzers is achieved and both
   can vary independently during program evolution.
   Examples of EAClusterAnalyzers are those that should return
   total energy of a cluster, the Center of Gravity, the Number of Cells. */
  virtual void modifyMe(EDClusterModifier & analyzer);
 
  //--
  ///  Dump for debugging.
  virtual void dump() const;

  //@}

  /**@name Status set/access functions */
  //@{
  /// Return a flag to tell that this EMClusterMember is NOT leaf .
  virtual bool isALeaf() const;

  /**
    Return the maximum energy measured in the cluster's (and sub clusters) 
    cells or the Energy of the cell itself. */
  virtual Energy_t  maxEnergy() const;


  //--
  /// Set the number of hypothetical particle that gave origin to this cluster.
  void setAllHypotheses(const ECClHyp & newhyp) { hyp = newhyp; }
  //--
  /// Get the number of all the hypotheses
  const ECClHyp & getAllHypotheses() const { return hyp; }

  //--
  /// Set the number of hypothetical particle that gave origin to this cluster.
  void setHypotheses(Hyp_t num) { hyp.setHypotheses(num); }
  //--
  ///  Get the number of hypotheses = (showers+charged particles).
  Hyp_t getHypotheses() const { return hyp.getHypotheses();}

  //--
  ///  Increase the number of linked tracks.
  Hyp_t newTrack() { 
    hyp.setTracks(hyp.getTracks()+1);
    return hyp.getTracks();
  };

  //--
  ///  Return the number of linked tracks.
  Hyp_t getTracks() const { return hyp.getTracks();}

  //--
  ///  Set the number of linked tracks.
  void setTracks(Hyp_t num) { hyp.setTracks(num);}

  //--
  ///  Set the number of Neutral Hypotheses.
  void setNeuHyp(Hyp_t num) { hyp.setNeuHyp(num);}
  //--
  ///  Get the number of Neutral Hypotheses.
  Hyp_t getNeuHyp() const { return hyp.getNeuHyp();}
  //--
  /// Set the number of Charged Hypotheses.
  void setChargHyp(Hyp_t num) { hyp.setChargHyp(num);}
  //--
  ///  Get the number of Charged Hypotheses.
  Hyp_t getChargHyp() const { return hyp.getChargHyp();}

  /// return the first child in the list.
  EMClusterMember * begin() const { return firstChild;}
  /// return the last child in the list.
  EMClusterMember * rbegin() const { return lastChild;}
  /// return end condition for a FORWARD iteration loop.
  EMClusterMember * end() const { return 0;}
  /// return end condition for a REVERSE iteration loop.
  EMClusterMember * rend() const { return 0;}
  //@}

  friend class EMClusterMember;
protected:
  /**@name Private Data */
  //@{
  ///  Pointer to the first child in the list.
  EMClusterMember*  firstChild;
  /// Pointer to the last child in the list.
  EMClusterMember*  lastChild;
  //@}
private:
  //--
  // /  Number of particle hypotheses in this cluster and below it. 
  //  int numHypot;
  //--
  // /  Number of stored extrapolated track points.
  //int numTracks;
  //--
  // /  Number of Neutral Hypotheses.
  //int numNeutr;
  //--
  // /  Number of Charged Hypotheses.
  //int numCharg;
  ECClHyp  hyp;
 
};


#endif // EMCLUSTER_HH_




