// *****************************************************************************
//
// source:    daq_mctruth_if.C
//
// type:      source code
//
// created:   14 Mar 2001
//
// author:    Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   functions to deliver fake FED IDs and contents
//            to wrap Monte Carlo truth information and deliver it
//            to the class DAQ_MCTRUTH_IF
//            called by DAQ_MC_IF::buildFED and DAQ_SPARSE_IF::setSLBs
//
// changes:   
// 03 Aug 2001 (BP) Added extra field to mimp information on the associated
//                  hits.
// 07 Aig 2001 (BP) Small bug correction.
//
// *****************************************************************************
#include "arte/ArteMessage.hh"

#include <stdlib.h>
#include <math.h>

#include "daqif/daq_mctruth_if.hh"
#include "daqif/kadaqif.hh"

#include "mctruth.h"


typedef hash_map<const char*, int, hash<const char*>, eqstr> hittab;

unsigned int DAQ_MCTRUTH_IF::getFED_IDs(unsigned short *buf, unsigned int maxIDs)
{
  unsigned int nfeds = 0; // return variable
  mctruth_feds.clear();
  if ( buf!=0 && ArteTable<MTRA>::size()!=0 )
   {
    for( ArteTable<MTRA>::iterator mtra = ArteTable<MTRA>::begin();
         mtra != ArteTable<MTRA>::end(); ++mtra) 
      {  
       int code = (*mtra)->topp->lund;  
       if (abs(code)==  11 // e +/-
        || abs(code)==  13 // mu +/-
        || abs(code)== 211 // pi +/-
        || abs(code)== 321 // K +/-
        || code== 2212     // proton
        || code==   22 )   // gamma
	{
         float Ekin = sqrt((*mtra)->px   * (*mtra)->px 
                         + (*mtra)->py   * (*mtra)->py 
                         + (*mtra)->pz   * (*mtra)->pz 
                         + (*mtra)->mass * (*mtra)->mass) - (*mtra)->mass;
         if (Ekin>=kadaqif::get_mctruth_Ecut())
	 {
          if (nfeds<maxIDs)
	   { 
             buf[nfeds] = mtra.index()-ArteTable<MTRA>::begin().index();
             mctruth_feds.insert(buf[nfeds]);
             nfeds++;
           }
	  else
	   {
	    arteMsg << "DAQ_MCTRUTH_IF::getFED_IDs: more than " << maxIDs << " tracks!" ;
            arteMsg.warn();
	    break;
           }
	 } // if (Ekin>=daqif_mctruth_Ecut)
	} // if ( (abs(code) ...
      } // for( ArteTable<MTRA> ...
   } // if (buf!=0 ...
  return nfeds;
}

  int DAQ_MCTRUTH_IF::getFED(unsigned short fedid, FED_HDR *buf, unsigned int maxSize)
{
  unsigned int length = sizeof(mtrack)+sizeof(FED_HDR)+sizeof(FED_TAIL);

  //check for space
  if (length <= maxSize && buf!=0)
    {
     mtrack *p = (mtrack*)(buf+1);

     ArteTable<MTRA>::iterator mtra = ArteTable<MTRA>::begin() + fedid;
  
     //Fill body of the data part

     p->orig.compid = 1; //used as version number!
     p->orig.nsl    = 0;
     p->orig.x      = (*mtra)->bver->x;
     p->orig.y      = (*mtra)->bver->y;
     p->orig.z      = (*mtra)->bver->z;
     if ((*mtra)->pz!=0)
      {
       p->orig.tx   = tan((*mtra)->px/(*mtra)->pz);
       p->orig.ty   = tan((*mtra)->py/(*mtra)->pz);
      }
     else
      {
       p->orig.tx   = 0;
       p->orig.ty   = 0;
      }
     p->orig.p      = sqrt((*mtra)->px * (*mtra)->px 
                          + (*mtra)->py * (*mtra)->py 
                          + (*mtra)->pz * (*mtra)->pz);
     if ((*mtra)->pz<0) p->orig.p = - p->orig.p;

     p->mass = (*mtra)->mass;
     if ( (*mtra)->topp->lund ==   11
       || (*mtra)->topp->lund ==   13
       || (*mtra)->topp->lund == -211
       || (*mtra)->topp->lund == -321 )
       p->mass = - p->mass;

     p->parentid=p->parentindex=p->parentvtx=0;

     if ((*mtra)->btra!=0) {
       p->parentid    = (*mtra)->btra->topp->lund;
       
       //Check if parent is among FEDs
       set<unsigned int>::iterator parent = mctruth_feds.find((*mtra)->btra.index()-ArteTable<MTRA>::begin().index());
       if (parent == mctruth_feds.end())
	 p->parentindex = -1;
       else
	 p->parentindex = (*mtra)->btra.index()-ArteTable<MTRA>::begin().index();
       
       if ((*mtra)->bver->flag<=1) p->parentvtx=1;
       else p->parentvtx=0;
     } 
     p->numpoints   = 0;

     //find out which MIMPs to store

     omt outmimps;
     hittab outhit;
     outmimps.clear();
     outhit.clear();
     
     for (ArteRelationIterator<MIMP> mimp=(*mtra)->mimp.begin();
          mimp!=(*mtra)->mimp.end(); ++mimp)
      {
	//Hitb counting...
	int nhits=0;
	for(ArteRelationIterator<HITB> hitb=(*mimp)->hitb.begin();
	    hitb!=(*mimp)->hitb.end(); ++hitb) nhits++;
	if (((*mimp)->gede->sin==0&&(*mimp)->gede->rot[0]!=1.)||(fabs((*mimp)->gede->sin)>0.2))
	  //stereos
	  nhits*=256;

       //Look up whether superlayer is already known for this track
       omt::iterator cur = outmimps.find((*mimp)->gede->gesl->name);
       //If superlayer is not already known, register it.
       if (cur==outmimps.end()) {
	 outmimps[(*mimp)->gede->gesl->name] = mimp;
	 outhit[(*mimp)->gede->gesl->name] = nhits;
       } else {
	 outhit[(*mimp)->gede->gesl->name] += nhits;
	 //If the new MIMP is before the registered MIMP, replace
	 if ((*mimp)->z<(*cur->second)->z)
	   outmimps[(*mimp)->gede->gesl->name] = mimp;
       }
      }

     //Fill mimps of the data part

     mtpoint *q=(mtpoint*)(p+1);

     for (omt::iterator om = outmimps.begin();
          om!= outmimps.end(); om++, q++, p->numpoints++)
      {
       length += sizeof(mtpoint);
       //check for space
       if (length <= maxSize)
	 {
	   q->compid = (*om->second)->gede->cmp;
	   q->nsl    = (*om->second)->gede->gesl->nsl;
	   q->x      = (*om->second)->x;
	   q->y      = (*om->second)->y;
	   q->z      = (*om->second)->z;
	   q->tx     = (*om->second)->tx;
	   q->ty     = (*om->second)->ty;
	   q->p      = (*om->second)->p;
	   q->hitb   = outhit[(*om->second)->gede->gesl->name];
	   //try without byte swap
	   //	   ESP_put_mimp(q,q);
	 }
       else
	 {
           arteMsg << "DAQ_MCTRUTH_IF::getFED: Too many MIMPS " << p->numpoints ;
	   //           cout << " These are all the MIMPs to be stored:" << endl;
           //unsigned short i=0;
           //for (omt::iterator dom = outmimps.begin();
           //     dom!= outmimps.end(); dom++, i++)
	   //  cout << i
           //       << ": cmp: " << (*dom->second)->gede->cmp
           //       <<  " nsl: " << (*dom->second)->gede->gesl->nsl
           //       <<    " z: " << (*dom->second)->z << endl;
           p->numpoints--;
 	   break;
	 }
      }

	   //try without byte swap
	   //     ESP_put_mtrack(p,p);

     //Build the FED record 

     //fill the header
     buf->length = length;
     buf->recid = RID_NORMAL_DEFAULT_SPARSE;
     buf->fltevt = 0;
     buf->bxid = 0;
     buf->compid = COMP_ID_MCTRUTH;
     buf->fedid = fedid;

     // fill the tail
     FED_TAIL *tail = (FED_TAIL*)(q+1);
     tail->fltevt = 0; 

     // swap if needed
     ESP_put_fed_hdr(buf, buf);
     ESP_put_fed_tail(tail, tail);
    }
  else 
    {
     arteMsg << "DAQ_MCTRUTH_IF::getFED: maxSize " << maxSize
             << " not sufficient or empty buf" ; 
     arteMsg.warn();
     return 2;
    }
  return 0;
}

int DAQ_MCTRUTH_IF::prepareEvent(void)
{
  return 1; // indicates that FED_IDs have to be re-read
}
