// *****************************************************************************
//
// source:    daq_otr_if.C
//
// type:      source code
//
// created:   13 Mar 2001
//
// author:    Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   functions to deliver FED IDs and contents to the class
//            DAQ_OTR_IF
//            called by DAQ_MC_IF::buildFED and DAQ_SPARSE_IF::setSLBs
//
// changes:   
//
// *****************************************************************************
#include "arte/ArteMessage.hh"

#include "daqif/daq_otr_if.hh"

extern "C" {
#include "esp.h"
	   }

unsigned int DAQ_OTR_IF::getFED_IDs(unsigned short *buf, unsigned int maxIDs)
{
  return MCGetFedIds(buf, maxIDs);
}

int DAQ_OTR_IF::getFED(unsigned short fedid, FED_HDR *buf, unsigned int maxSize)
{
  int rc = 0;

  //Look up the FED.

  OTRmap_t::iterator curf = OTRmap.find(fedid) ;

  unsigned int length;
  unsigned short odd = 0;
  if (curf!=OTRmap.end()) 
    {
     odd = curf->second.size() % 2;
     length = 2*(curf->second.size()+odd)+sizeof(FED_HDR)+sizeof(FED_TAIL);
    }
  else
    {
     length = sizeof(FED_HDR)+sizeof(FED_TAIL);
    }
  //check for space
  if (length <= maxSize && buf!=0)
    {
     //Build the FED record 

     //fill the header
     buf->length = length;
     buf->recid = RID_NORMAL_DEFAULT_SPARSE;
     buf->fltevt = 0;
     buf->bxid = 0;
     buf->compid = COMP_ID_OTR;
     buf->fedid = fedid;

     unsigned char *p = (unsigned char*)(buf+1);

     if (curf!=OTRmap.end()) 
       {
        // fill the data
        for ( OTRFED_t::iterator curc=curf->second.begin();
              curc != curf->second.end(); curc++, p++, p++ )
	  {
           *p = curc->first;
           *(p+1) = curc->second;
	  } // for ( OTRFED_t::...
        if (odd)
          {
           *p = (char)0xFF;
           *(p+1) = (char)0xFF;
           p++;p++;
          }
       }
     // fill the tail
     FED_TAIL *tail = (FED_TAIL*)p;
     tail->fltevt = 0; 
     // swap if needed
     ESP_put_fed_hdr(buf, buf);
     ESP_put_fed_tail(tail, tail);
    }
  else 
    {
      arteMsg << "DAQ_OTR_IF::getFED: maxSize " << maxSize << " not sufficient or empty buf" ;
      arteMsg.warn();
      rc=2;
    }
  return rc;
}

int DAQ_OTR_IF::prepareEvent(void)
{
  int rc=0; // 0 means FED_IDs will not change

  OTRmap.clear();

  unsigned short fedid; 
  unsigned char chanid;
  unsigned char tdccount;

  // loop over all OTR hits

  for ( ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin();
        hitbIt != ArteTable<HITB>::end(); hitbIt++ )
    {
      ArtePointer<HITB> hitb = *hitbIt;
      if (hitb->gede->cmp!=2) continue; //according to auto/all.kumac: ' 2 WCH wire chambers'
      int status=MCGetFedInfo(*hitbIt, fedid, chanid, tdccount);
      if (status!=0) 
	{
	  arteMsg << "MCGetFedInfo returned " << status; // not a severe error
	  arteMsg.warn();
          break;
        }
      else
        {  
	  //Look up whether fedid is already known.
          OTRmap_t::iterator cur = OTRmap.find(fedid);
          //If fedid is not already known, register it.
          if (cur==OTRmap.end()) 
            {
	     pair<OTRmap_t::iterator,bool> fins = 
              OTRmap.insert(pair<const unsigned short, OTRFED_t>(fedid,OTRFED_t()));
             if (fins.second)
               {
                cur = fins.first;
		// insert initial hit
                pair<OTRFED_t::iterator,bool> cins = 
                 cur->second.insert(pair<const unsigned char, unsigned char>(chanid,tdccount));
		if (!cins.second) 
                  {
                   arteMsg << "DAQ_OTR_IF::getFED_IDs: OTRFED_t.insert failed" ;
                   arteMsg.warn();
                   rc=-1;
                  }
               }
             else 
               {
                arteMsg << "DAQ_OTR_IF::getFED_IDs: OTRmap.insert failed" ;
                arteMsg.warn();
                rc=-2;
               }
            }
          else
            {
             //Look up whether chanid of this fedid is already known.
             OTRFED_t::iterator chan = cur->second.find(chanid);
             //If chanid is not already known, register it.
             if (chan == cur->second.end()) 
               {
                pair<OTRFED_t::iterator,bool> cins = 
                 cur->second.insert(pair<const unsigned char, unsigned char>(chanid,tdccount));
		if (!cins.second) 
                  {
                   arteMsg << "DAQ_OTR_IF::getFED_IDs: OTRFED_t.insert failed" << endl;
                   arteMsg.warn();
                   rc=-3;
               }
             else 
               {
                //If known, compare the TDCs, store the smaller TDC.
                if (chan->second > tdccount) chan->second = tdccount;
               }
               }
	    } //cur==OTRmap.end
	} //status!=0
    }  //for ( ArteTable...

  return rc;
}
