// *****************************************************************************
//
// source:    daqif.C
//
// type:      source code
//
// created:   14 Mar 2001
//
// author:    Mike Medinnis, Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   generating daq-format data from monte carlo
//
// changes:   
//    
// 30 Jul 2001 (BP) bug correction for SLTsim
//
// *****************************************************************************
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"

#include <arte/ArteProto.hh>

#include "daqif/daq_mc_if.hh"
#include "daqif/daq_daq_if.hh"
#include "daqif/daq_targ_if.hh"
#include "daqif/daq_rich_if.hh"
#include "daqif/daq_ecal_if.hh"
#include "daqif/daq_svd_if.hh"
#include "daqif/daq_otr_if.hh"
#include "daqif/daq_muon_if.hh"
//#include "daq_mpix_if.hh"
#include "daqif/daq_mpad_if.hh"
#include "daqif/daq_itr_if.hh"
#include "daqif/daq_mctruth_if.hh"
#include "sltarte/sltsim.hh"

extern "C" {
#include "esp.h"
	   }


int DAQ_MC_IF::nInstances[COMP_NUMBER] = {0};
DAQ_MC_IF *DAQ_MC_IF::ifs[COMP_NUMBER] = {0};

char DAQ_MC_IF::fedBuf[MAX_FED_SIZE];

int DAQ_MC_IF::fd = -1;
unsigned int DAQ_MC_IF::bufLen = 0;
char *DAQ_MC_IF::evtBuf = 0;

int DAQ_MC_IF::compMask = 0;
bool DAQ_MC_IF::initialized = false;

static int fltevt=0;

/***
   * Initialize must be called before anything else. 
   *
   ***/

int DAQ_MC_IF::initialize(unsigned int comp_mask)
{
  if (initialized) {
    arteMsg << "Tried to initialize DAQ_MC_IF twice.";
    arteMsg.warn();
    return 1;
  } else {
    compMask = comp_mask;
    init_interfaces();
    initialized=true;
    return 0;
  }
}

/***
   * initialize the pipe to EDC
   *
   ***/

int DAQ_MC_IF::init_pipe(void)
{
  if (fd != -1) return 1;
  int rc = 0;

  int pfd[2];
  if (pipe(pfd)==-1) {
    arteMsg << "daqif: error creating pipe" ;
    arteMsg.warn();
    return 1;
  }
  int stat=fork_SMC();
  if (stat==0) rc = fork_EDC(pfd);
  else rc = stat;
  if (std::close(pfd[0])==-1) {
    arteMsg << "daqif: error closing pfd[0]";
    arteMsg.warn();
  }

  fd=pfd[1];

  return rc;
}

/***
   * initialize the output file
   *
   ***/

int DAQ_MC_IF::init_file(char *filename)
{
  if(fd != -1) return 1;
  int rc = 0;
  fd = open(filename,O_WRONLY|O_CREAT|O_TRUNC,0666);
  if(fd == -1) {
    rc=3;
  }
  return rc;
}


/***
   * instantiate all existing interfaces, even if masked
   * compMask will be applied in buildEvent
   *
   ***/

void DAQ_MC_IF::init_interfaces(void)
{
  new DAQ_DAQ_IF();

  new DAQ_TARG_IF();
  new DAQ_ECAL_IF();
  new DAQ_RICH_IF(); 

  new DAQ_SVD_IF();
  new DAQ_OTR_IF();

  new DAQ_MUON_IF();
  //  new DAQ_MPIX_IF(); // not yet existing
  new DAQ_MPAD_IF();

  new DAQ_ITR_IF();

  new DAQ_MCTRUTH_IF();

  return;
}


/***
   * build the daq event in memory
   * return a pointer to it and its length in *nBytes
   *
   ***/

EVENT_HDR* DAQ_MC_IF::getEvent(unsigned int *nBytes)
{
  *nBytes = buildEvent(evtBuf, bufLen);

  // If the event is too large for the buffer (guaranteed the first time around),
  // try to allocate a larger buffer
  if(*nBytes > bufLen) {
    bufLen = *nBytes + *nBytes/16; // a little fat
    // cerr << "DAQ_MC_IF: allocating " << bufLen << " bytes for event building." << endl;
    if(evtBuf != 0)
      free(evtBuf);
    evtBuf = (char*) malloc(bufLen);
    if(evtBuf == 0) {
      arteMsg << "DAQ_MC_IF::getEvent attempt to allocate " << bufLen 
              << " bytes failed, event not written" ;
      arteMsg.warn();
      return 0;
    }
    if( (*nBytes = buildEvent(evtBuf, bufLen)) > bufLen) {
      arteMsg << "DAQ_MC_IF::getEvent second attempt to build event failed ... fire the programmer";
      arteMsg.warn();
      return 0;
    }
  }
  return (EVENT_HDR*)evtBuf;
}


/***
   * write the daq event to the file descriptor fd (pipe or file)
   *
   ***/

int DAQ_MC_IF::writeEvent(EVENT_HDR* evtBuf)
{
  if(fd == -1 ) {
    arteMsg << "DAQ_MC_IF: attempt to write without having initialized" ;
    arteMsg.warn();
    return 1;
  }
  int nBytes=evtBuf->length ;
  ESP_get_int(&nBytes,&nBytes,1);

  size_t nWritten = write(fd, evtBuf, nBytes);
  if(nWritten != nBytes)
    return 4;

  return 0;
}


/***
   * The main task of the constructor is to keep a pointer to the instance under 
   * construction -- which will be used by getDAQ_IF.
   *
   * Note that because virtual functions are used, it is essential that the 
   * instances of subclasses (one per component) are allocated on the heap
   * (via the new operator). This is best done in the initialization for the
   * component --- umm, assuming it is written in C++.
   *
   * Note that "component_id" is a number between 0 and COMP_NUMBER -- the number
   * in the various headers is the same number left-shifted by 8.
   *
   ***/

DAQ_MC_IF::DAQ_MC_IF(int component_id = -1) : compid(component_id), nSLBs(0), setSLBsCalled(0) 
{
  if(component_id < 0 || component_id >= COMP_NUMBER) {
    arteMsg << "DAQ_MC_IF:: component id out of range " << component_id << endl;
    arteMsg.warn();
    return;
  }
  if(++nInstances[component_id] != 1) {
    arteMsg << "DAQ_MC_IF: warning more than one instance for component " << component_id;
    arteMsg.warn();
  }
  ifs[component_id] = this;

}

/***
   * Return a pointer to the interface object of component_id, if it exists.
   * Otherwise, return 0.
   *
   ***/

const DAQ_MC_IF *DAQ_MC_IF::getDAQ_IF(int component_id)
{
  if(component_id < 0 || component_id >= COMP_NUMBER)
    return 0;
  return ifs[component_id];
}

/***
   * Return the FEDs of a given SLB
   *
   ***/

unsigned short* DAQ_MC_IF::getFEDs_for_SLB(int slbID, int *n_feds)
{
  if(slbID < 0 || slbID >= nSLBs) {
    *n_feds = 0;
    return 0;
  }
  *n_feds = SLB[slbID].nFEDs;
  return SLB[slbID].first;
}

/***
   * Build an event in buf. If there isn't enough room, the event is not
   * correctly built (but shouldn't result in writing beyond the end of
   * buffer). In any case the value returned is the length of event. The
   * idea is that the user checks if the return value is greater than
   * maxSize, and if so, reallocate a buffer large enough to hold the
   * event.
   *
   ***/

unsigned int DAQ_MC_IF::buildEvent(char *buf, unsigned int maxSize)
{
  unsigned int nBytes = 0, i;
  ArtePointer<EVHD> evhd(1);
  int sltevt = evhd->evt;

  fltevt++;
  if (fltevt>127) fltevt=0;
    
  nBytes += sizeof(EVENT_HDR);
  EVENT_HDR *evt = (EVENT_HDR *) buf;

  if(nBytes <= maxSize) {
    evt->recid = RID_NORMAL_EVENT;
    evt->format = ESP_big_endian; // since it's going to be...
    evt->sltevt = sltevt;
    evt->compmask = 0;
    evt->evtmask = 1;
  }

  maxSize = ( nBytes < maxSize)? maxSize - nBytes : 0;
  int c_mask = compMask;
  for(i = 0; c_mask != 0; i++, c_mask >>= 1) {
    if( (c_mask & 1) && ifs[i] != 0) {
      if(nBytes <= maxSize)
	evt->compmask |= (1 << i);
      unsigned int compBytes = ifs[i]->buildComponent(buf + nBytes, maxSize);
      maxSize = ( compBytes < maxSize)? maxSize - compBytes : 0;
      nBytes += compBytes;
    }
  }

  if( sizeof(EVENT_TAIL) <= maxSize) {
    EVENT_TAIL *tail = (EVENT_TAIL*) (buf + nBytes);
    tail->sltevt = sltevt;
    evt->length = nBytes + sizeof(EVENT_TAIL);
    ESP_put_event_hdr (evt, evt);
    ESP_put_event_tail(tail, tail);
  }

  nBytes += sizeof(EVENT_TAIL);
  return nBytes;
}

/***
   * Build a component. Return the total length of this component. 
   *
   ***/

unsigned int DAQ_MC_IF::buildComponent(char *buf, unsigned int maxSize)
{
  unsigned int nBytes = 0;
  int status = prepare();
  if (status != 0) 
    {
      arteMsg << "ERROR: prepare failed for component " << compid 
              << " with status return " << status ;
      arteMsg.warn();
      nSLBs = 0;
    }
  if(nSLBs == 0)
    return 0;

  for(int id = 0; id < nSLBs; id++) {
    unsigned int compBytes = buildSLB(id, buf + nBytes, maxSize);
    maxSize = (compBytes < maxSize)? maxSize - compBytes : 0;
    nBytes += compBytes;
  }
  return nBytes;
}

/***
   * Build one SLB in the space provided (buf).
   * slbid's are numbered consecutively starting
   * from zero. For the forseeable future, it's enough that each component has all FEDs
   * in a single SLB. Return the total length. If there isn't enough room in buf, stop
   * writing to it but continue to compute the length.
   *
   ***/

unsigned int DAQ_MC_IF::buildSLB(unsigned int slbid, char *buf, unsigned int maxSize)
{
  int nGoodFeds = 0;

  // leave some room for the SLB header ...
  SLB_HDR *hdr = (SLB_HDR*) buf;
  unsigned int nBytes = sizeof(SLB_HDR);
  
  maxSize = (nBytes < maxSize)? maxSize - nBytes : 0; 

  int n_feds;
  unsigned short *ids = getFEDs_for_SLB(slbid, &n_feds);

  for(int fed = 0; fed < n_feds; fed++) {
    FED_HDR *anFed = (FED_HDR*) &fedBuf;
    int status = buildFED(compid, ids[fed], anFed, MAX_FED_SIZE);
    if(status != 0) {
      arteMsg << "DAQ_MC_IF: error code " << status 
              << " returned from buildFED for compid: : " << compid ;
      arteMsg.warn();
      continue;
    }
    ESP_get_fed_hdr(anFed, anFed);
    unsigned int length = anFed->length 
                        - sizeof(FED_HDR) - sizeof(FED_TAIL) + sizeof(SLB_SUBHDR);
    if(length > maxSize) {
      nBytes += length;
    } else {

      // Make an SLB subheader then copy the data portion
      SLB_SUBHDR *sub = (SLB_SUBHDR*) (buf + nBytes);
      sub->length = length;
      sub->fedid  = anFed->fedid;
      sub->recid  = anFed->recid;
      sub->bxid   = anFed->bxid;
      ESP_put_slb_subhdr(sub, sub);
      nBytes += sizeof(SLB_SUBHDR);
      memcpy( buf + nBytes,
              anFed + 1,
              anFed->length - sizeof(FED_HDR) - sizeof(FED_TAIL));
      nBytes += (length - sizeof(SLB_SUBHDR));

      // build SLB header & trailer (except length) from first valid fed
      if( ++nGoodFeds == 1) {
	hdr->recid =   RID_NORMAL_SLB0;
	hdr->fltevt = anFed->fltevt;
	hdr->compid = anFed->compid;
	hdr->slbid = slbid;
	hdr->nodeid = 0;
      }
    }
    maxSize = (length < maxSize)? maxSize - length : 0;
  }  // end of: for(int fed = 0; fed < n_feds; fed++)

  if( sizeof(SLB_TAIL) <= maxSize ) {
    if(nGoodFeds == 0) {
      hdr->recid =  0;
      hdr->fltevt = 0;
      hdr->compid = compid << 8;
      hdr->slbid = slbid;
      hdr->nodeid = 0;
    }
    hdr->length = nBytes + sizeof(SLB_TAIL);
    SLB_TAIL *tail = (SLB_TAIL*) (buf + nBytes);
    tail->recid = hdr->recid;
    tail->fltevt = hdr->fltevt;
    ESP_put_slb_hdr(hdr, hdr);
    ESP_put_slb_tail(tail, tail);
  }
  nBytes += sizeof(SLB_TAIL);
  return nBytes;
}

/***
   * return the value for the field fltevt
   *
   ***/

short DAQ_MC_IF::get_fltevt(void)
{
 return fltevt;
// ArtePointer<EVHD> evhd(1);
// return evhd->evt & 0x7F;
}

/***
   * Use the getFED function to pull in the "raw" FED -- then 
   * fix up the common words: bxid, fltevt
   *
   ***/

int DAQ_MC_IF::buildFED(unsigned int component, unsigned short fedid, FED_HDR *buf, unsigned int maxSize)
{
  if(component < 0 || component > COMP_NUMBER)
    return 1;
  if( ifs[component] == 0)
    return 2;
  int status = ifs[component]->getFED(fedid, buf, maxSize);
  if(status != 0)
    return status + 2;
  ESP_get_fed_hdr(buf, buf);
  buf->fltevt = get_fltevt();
  buf->bxid = buf->fltevt;
  FED_TAIL *aTail = (FED_TAIL*) ( (char*) (buf + buf->length - sizeof(FED_TAIL)) );
  aTail->fltevt = buf->fltevt;
  ESP_put_fed_hdr(buf, buf);
  ESP_put_fed_tail(aTail, aTail);
  return 0;
}
