// *****************************************************************************
//
// source:    fillITRSEVTfromHITB.C
//
// type:      source code
//
// created:   14 Mar 2001
//
// author:    B. Schwingenheuer, Brian Petersen, Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   fill Marvin Sparse Event Buffer from HITB for ITR
//            called by daq_svd_if.C
//
// changes:   
// 02-aug-2001 (BP) Reduce maximum size to avoid overflows.
//
// *****************************************************************************
#include "arte/ArteMessage.hh"
#include "arte/arte.hh"
#include "ITR_marvin.h"
#include "itrleda.h"

// memory alleocation for ITR Marvin Sparse Event: need 9 bytes/cluster
// 45000 bytes correspond to 5000 clusters
#define SEVT_INIT_SIZE 31520 
#define MAXHIT ((SEVT_INIT_SIZE-20)/9)
extern "C" {int ItrInit();}

int fillITRSEVTfromHITB(int major, int minor)
{
  static ITR_Marvin_Sevt_t  *sevt;

  int celllong = 0;
  int cellshort = 0;
  int format = 3;
  int ncl = 0;
  int width = 2;

  int isize = sizeof(int);
  int ssize = sizeof(short int);
  int csize = sizeof(char);

  float  cog, sn, charge,relpos;
  float  f4sn;
  unsigned char bcharge, b4sn;
  signed char   bwidth;
  short int schamber;

  static int first_marvin = 1;

  char *routine="fillITRSEVTfromHITB";

  /* initialize marvin and allocate some space for sparsified event */
  if ( first_marvin ) {
    ItrInit(); // Load Leda tables - not normally used in MC unless reco is on.
    ITR_marvin_init();

    sevt = ITR_marvin_fetch_sevt();
    if ( !sevt ) {
      arteMsg << routine
              << "no Marvin_Sevt_t available (marvin not initialized?)";
      arteMsg.warn();
      return 1;
    }

    sevt->data   = (char*) malloc(SEVT_INIT_SIZE); 
    sevt->size   = SEVT_INIT_SIZE;
    sevt->format = format;

    first_marvin = 0;
 
  }
 
  sevt->next = sevt->data;

  /* initialize an sevt-header */
  ESP_put_int( &major, (int*) sevt->next, 1);
  sevt->next += isize;
  ESP_put_int( &minor, (int*) sevt->next, 1); 
  sevt->next += isize;
  ESP_put_int( &cellshort,  (int*) sevt->next, 1); 
  sevt->next += isize;
  ESP_put_int( &celllong,  (int*) sevt->next, 1); 
  sevt->next += isize;
  ESP_put_int( &format, (int*) sevt->next, 1); 
  sevt->next += isize;
  ESP_put_int( &ncl,    (int*) sevt->next, 1); 
  sevt->next += isize;


  // loop over all ITR hits, store them im sparse event format 3

  for ( ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin();
        hitbIt != ArteTable<HITB>::end(); hitbIt++ ) {
    ArtePointer<HITB> hitb = *hitbIt;

    ArtePointer<GEDE> gede = hitb->gede;

    ArtePointer<GEDE> itrgede = hitb->gede;

   
    // process ITR hits only 

    if(gede->cmp != 3) continue;
    // check for enough space in Sparse Event buffer 
    if(ncl+1 > MAXHIT) {
      arteMsg << routine
              << " Sparse Event Buffer full!";
      arteMsg.warn();
      return 0;
    }

    // find the chamber id:
    ArtePointer<GESL> gesl = gede->gesl;
    int sec=gede->sec;
    // There are two gede's for one nede)
    if (sec==2||sec==4) {
      itrgede=ArtePointer<GEDE> (itrgede.index()+1);
    }
    if (sec==6||sec==8) {
      sec--;
      itrgede=ArtePointer<GEDE> (itrgede.index()-1);
    }
    if (sec==1||sec==3) {
      sec++;
    }
  
    LedaPointer<class ITR_nede> nede = ITR_nede->get->key(gesl->nsl,
							  gede->lay,
							  sec);
    if (nede==NULL) {
      arteMsg << "No ITR nede found: nsl="
              << gesl->nsl << " lay: " << gede->lay << " sec:" << gede->sec;
      arteMsg.warn();
      continue;
    }
    LedaPointer<class ITR_Chamber> chamber = nede->tochamber;
    schamber=chamber->id;

    // Doesn't work as data has different system
    //    cog=(hitb->x -  gede->sxmn)/gede->pit;

    // These should work for map.c as was on Dec 18-2000

    if (sec<5)  // +x
      cog=(hitb->x -  itrgede->xfw[0])/gede->pit+8.0;
    else {      // -x
      if (gesl->nsl<9) {// MS01-MS13 has 752 wires - should take it from gede
	cog=752+7-(hitb->x -  itrgede->xfw[0])/gede->pit;
      } else
	cog=758+7-(hitb->x -  itrgede->xfw[0])/gede->pit;
    }
    //        printf("nsl=%d,lay=%d sec=%d rot=%f cog=%f xfw=%f pit=%f\n",gesl->nsl,gede->lay,sec,hitb->x,cog,itrgede->xfw[0],gede->pit);
    sn     = hitb->y;
    charge = hitb->q;
 
    f4sn = 4.*sn;
    if(charge>0) {
      bcharge = charge > 254. ? 255 : (int)charge;
      bwidth  = width  > 127  ? 127 : width;
      b4sn    = f4sn   > 254. ? 255 : (int)f4sn;
    } 
    else {
      bcharge = -charge > 254. ?  255 : (int)(-charge);             
      bwidth  =  width  > 127  ? -127 : -width;
      b4sn    = -f4sn   > 254. ?  255 : (int)(-f4sn);
    }
    ESP_put_short((short int *) (&schamber),(short int *) sevt->next,1);
    sevt->next += ssize;
    sevt->next = (char *)memcpy(sevt->next, &bcharge, csize) + csize;       
    sevt->next = (char *)memcpy(sevt->next, &b4sn   , csize) + csize;
    sevt->next = (char *)memcpy(sevt->next, &bwidth , csize) + csize;
    ESP_put_int( (int*) (&cog), (int*) sevt->next, 1);
    sevt->next += isize;
    ncl++;
   }

  // update cluster count in sparse event header
  ESP_put_int( &ncl, (int*) sevt->data + 5, 1);
  return(0);
}
