/************************************************************************
* This routine translates GEANT output hits into ARTE tables
* HITB and DIGB for OTR modular geometry.
*
* Called from DGEVNT.
*
* Written by S.Nowak, C.Stegmann, A.Lanyov     Oct 1999.
*
* Changes: 
* 11/11/2000: implementation of OtrI into the hit digitization 
*
* Arguments  kind         description
*-----------------------------------------------------------------------
* IPL        in/out       Detector number in GEDE bank
************************************************************************/
#include <vector>
#include <iostream>
#include <algorithm>
#include <math.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include "gean/ZGESL.h"
#include "digi/zgdig.hh"
#include "digi/dglev.h"
#include "digi/dgnoise.h"
#include "digi/dgocell.hh"
#include "Otr/OtrI.h"
#include "Otr/OtrMc.h"

extern "C" {
  void dgomod_(int &);
  void grndm_(float [], const int &);
  void granor_(float &, float &);
  void gpoiss_(float *,int * ,int &);
  void sortzv_(int[], int[], int&, int&, int&, int&);
  int otr_gede_status(int, int, int);
}

inline int OtrEventChannelValid(int ci);

// routine used to sort the vector with hits in z and x
inline bool compareDIGX(const digx& lhs, const digx& rhs) {
  return lhs.strip < rhs.strip ;
}

void dgomod_(int & ipl) {
  
  static int hitprepflag=-1;
  int i1=1 ; 
  float ydg10 = 2.5, rndm[3];
  vector<digx> *dg_otr[4];
  for(int i=0; i<4; i++) dg_otr[i] = 0 ;
  
  //==================================================================
  //  Switch between WAX and OtrI for maskfile treatment    UU 
  //==================================================================
  
  OtrMcKillStatistics("Tstart", 0);
  
  if (hitprepflag==-1) {
    char *f;
    int ci, asd, cha;
    int dead=0, noisy=0;

/*  check which hit preparation is selected for OTR  */
    if ((f=getenv("OtrHitPreparation"))) hitprepflag = atoi(f);
    else hitprepflag=0;
    if (hitprepflag!=1) {
      printf("dgomod> OtrDigitization: OtrI used for dead channel killing.\n");
      if(dglev_.idglev[LGESL_WCH-1] == 3) {
	for (asd=0; asd<OtrIAsdNumberOf(); asd++){
	  for (cha=0; cha<OtrIAsdNumberOfChannel(); cha++) {
	    ci = OtrIFromAsd(asd, cha);
	    if (OtrIIsDead(ci)) dead++; 
	    else if (OtrIIsBad(ci)) noisy++; 
	  }
	} 
	printf("dgomod> OtrDigitization: OtrI Noisy/Dead Channels = %d %d \n\n",
	       noisy, dead);
      } else {
	printf("dgomod> OtrDigitization: --- dead cell killing not selected ---\n");
      }
    } else {
      printf("dgomod> OtrDigitization: WAX used for dead channel killing.\n");
      if(dglev_.idglev[LGESL_WCH-1] != 3) 
	printf("dgomod> OtrDigitization: --- dead cell killing not selected ---\n");
    }
  }
//--------------------------------------------------------------------------
 

   /* In GEAN96 all MIMP's of a module are assigned to the first GEDE of this 
   module. Therefore one has to calculate the number of GEDE's belonging 
   to this module (nsec) */
   ArtePointer<GEDE> gd(ipl);
   unsigned int j=0;
   for(j=ipl+1;;j++) {
     if(j > ArteTable<GEDE>::size()) break; 
     if(ArtePointer<GEDE>(j)->gcop != gd->gcop ) break;  
   }
   int nsec = j - ipl;
   if(nsec > 4) { 
     arteMsg << "dgomod: in GEDE plane "<<ipl<<" there are " <<nsec<<" sectors";
     arteMsg.error();
   }

   // define some often used quantitites identical for all sublayers
   zgdig ZGDIG(ipl);
   float rout =  gd->pit/3.;              // cell''s outer radius
   float rin  =  0.8660254*rout;          // cell''s inner radius
   
   // calculate number of mimps 
   int nmim=0, ifmimp=0;
   nmim = ZGDIG.Get_nmimp(ifmimp);
   
   // prepare intermediate storage
   digx otrdg;
   for(int i=0; i<nsec; i++) dg_otr[i] = new vector<digx>;

   /*The x coordinate will be
     transformed into the plane zmin of the given detector and rotated */
   
   for(int ih=0;ih<nmim;ih++) {  // loop over all MIMPs
      ArtePointer<MIMP> mp(ih+ifmimp);
      
      // rotate mimp into wire orientation
      float x_rot  = gd->rot[0]*mp->x + gd->rot[3]*mp->y ;
      float y_rot  = gd->rot[1]*mp->x + gd->rot[4]*mp->y ;
      float tx_rot = gd->rot[0]*mp->tx + gd->rot[3]*mp->ty ;
      float ty_rot = gd->rot[1]*mp->tx + gd->rot[4]*mp->ty ;
      
      float cx_rot = 0.;
      if(fabs(tx_rot) < 1.e+10) cx_rot = 1./sqrt(1.+ tx_rot*tx_rot);
      
      // calculate maximum and minimum y coordinate in rotated plane
      float y_min = y_rot + ty_rot* (gd->zmin - mp->z); 
      float y_max = y_rot + ty_rot* (gd->zmax - mp->z);
      if(y_min > y_max) swap(y_min,y_max) ;
      
      // check into which sector the mimp goes
      for(int j=0; j<nsec; j++) { // loop over all sectors (gedes)
	ArtePointer<GEDE> gf(ipl+j);

	// check if mimp hits module in y
	if( y_min < gf->symx && y_max > gf->symn ) {
	  
	  for(int itube=0;itube<gf->nsl;itube++)   
	    if( gf->nwi[itube] > 0 ) {   // loop over monolayers =tubes
	      
	      // strange correction for dead region from g10 strips
	      float y1w = gf->y1w[itube] + 0.5 * ydg10;
	      float y2w = gf->y2w[itube] - 0.5 * ydg10;

	      // calcualte mimp-position at z-wire:
	      float wz     = gf->zfw[itube];
	      float wx_rot = x_rot + tx_rot*(wz - mp->z) ;
	      float wy_rot = y_rot + ty_rot*(wz - mp->z) ;
		
	      // require that track with y limit at z-position of cell
	      // (such that it is at least half in the cell)	      
	      if( y1w < wy_rot && wy_rot < y2w) {

		// x position at borders of monolayer:
		float wx_zmin = wx_rot - rin*tx_rot ;
		float wx_zmax = wx_rot + rin*tx_rot ;
		
		// first/last cell traversed by track (counting from 1!)
		float xfw = gf->xfw[itube] ;
		int nw1 = int( (wx_zmin - ( xfw - gf->pit/2 ) )/gf->pit) + 1;
		int nw2 = int( (wx_zmax - ( xfw - gf->pit/2 ) )/gf->pit) + 1;
		if(nw1>nw2) swap(nw1,nw2) ;
		if(nw1<1)   nw1 = 1;
		if(nw2>gf->nwi[itube]) nw2 = gf->nwi[itube] ;
		
		for (int n=nw1;n<=nw2;n++) {  // loop over wires
		  
		  // Check status of wire and ignore the hit if this wire is bad.
		  if(dglev_.idglev[LGESL_WCH-1] >= 3) {  // dglev OTR 3
		    int status = 0; 
		    if (hitprepflag==1) {
		      status = otr_gede_status(ipl+j, itube, n-1);
		    } else {
		      int ci = OtrIFromNedeLayerWire(gf.index(), itube, n-1);
		      // int ci = OtrIFromHwcLayerWire(gf->hwc, itube, n-1);
		      // status = OtrIIsDead(ci); 
		      status = !OtrEventChannelValid(ci);
		    }
		    OtrMcKillStatistics("Add", status);
		    if (status) continue;
		  }
		  
		  float xwir = xfw + (float(n)-1.)*gf->pit; // x-position of wire
		  float dd = (wx_rot-xwir) * cx_rot; 	    // smallest distance to wires
		  float dpath =                             // path length in cell 
		    dgocell::PathLengthHC(gf->pit,wx_rot-xwir,tx_rot,ty_rot) ;
		  
		  // track hits cell
		  if(dpath > 0 ) {
		    if(dglev_.idglev[LGESL_WCH-1] != 0) {    //long version
		      otrdg.charge = 
			dgocell::CreateHit(gf->pit,gf->acc,gf->eff,dpath,dd ) ;
		      if ( otrdg.charge > 0 ) {          // hit accepted
			otrdg.strip = (itube<<16) + n;   // strip number
			otrdg.mimp  = mp.index();        // mimp number
			dg_otr[j]->push_back(otrdg);
		      } 
		    } else  {  //short version
		      grndm_(rndm,i1);
		      if(rndm[0] > gf->eff) continue;
		      granor_(rndm[0],rndm[1]);
		      dd = dd + gd->acc*rndm[0]; 
		      if (fabs(dd) > rout) continue;
		      if(fabs(dd) > gf->dod) continue; 
		      otrdg.strip = (itube<<16) + n;             // strip number
		      otrdg.charge= dd ;//+ gd->acc*rndm[0];     // distance
		      otrdg.mimp = mp.index();                   // mimp number
		      dg_otr[j]->push_back(otrdg);
		    }
		  }
		}    //======================== end of loop on wires
	      }    
            }        //======================== end of loop on monolayers
	} 
      }             //======================== end of loop on sectors 
   }                //======================== end of loop on mimps
   
   // generate X talk and noise
   if(dglev_.idglev[LGESL_WCH-1] != 0)  //long version
     for (int j = 0; j< nsec; j++) {
       ArtePointer<GEDE> gf(ipl+j);
       dgocell::GenerateXTalkHits(*(dg_otr[j]),gf) ;
       dgocell::GenerateNoiseHits(*(dg_otr[j]),gf) ;
     }
   
   for (int j=0;j< nsec;j++) {  //======================== loop on sectors
     
      int ndigde = dg_otr[j]->size();
      ArtePointer<GEDE> gf(ipl+j);

      if (ndigde == 0) {
	gf->dig1 = 0 ; gf->hit1 = 0; 
	gf->digl = -1; gf->hitl = -1;  
      } else {
      
	// for backwards compatibillity the digx are sorted in
	// increasing tube/wire number. I do not think that this 
	// is really needed, but we shall make maximum use of it ... (WH)
	sort(dg_otr[j]->begin(),dg_otr[j]->end(),compareDIGX) ;

	ArtePointer<DIGB> digb = ArteTable<DIGB>::allocate(dg_otr[j]->size()) ;
	gf->dig1 = digb ;
	int prevstrip = -1;
	ArtePointer<HITB> hitb ;
	for(vector<digx>::iterator digxIt = dg_otr[j]->begin() ;
	    digxIt != dg_otr[j]->end() ; digxIt++, digb++) {
	  gf->digl = digb ;
	  
	  int itube = digxIt->strip >> 16;
	  int iwire = digxIt->strip&65535;
	  digb->id = 10000*((iwire-1)*4 + itube+1) + ipl + j;
	  //      id = 111 0068 means strip=28, tube=3, ipl+j=68
	  digb->q  = digxIt->charge ; 
	  if (digxIt->mimp != 0) 
	    createRelation(ArtePointer<MIMP>(digxIt->mimp),digb,&MIMP::digb) ;
	  
	  // create new hitb entry if not existent for this strip
	  // (this only works if digx are sorted)
	  if(prevstrip != digxIt->strip) {
	    // book new entry in hitb bank
	    bool firsthitb = hitb.index()==0 ;
	    hitb = ArteTable<HITB>::push_back(HITB()) ;
	    if(firsthitb) gf->hit1  = hitb ;
	    gf->hitl = hitb ;
	    
	    hitb->gede = ipl + j;
	    unsigned char mono = itube ;
	    unsigned char cell = iwire-1 ;

	    hitb->x  = gf->xfw[mono] + cell * gf->pit;
	    hitb->z  = gf->zfw[mono];
	    if (hitprepflag==1) {
	      // wax definition
	      hitb->chid = cell + 32*mono ;
	    } else {
	      // otri definition
	      int ci=OtrIFromNedeLayerWire(hitb->gede.index(),mono,cell);        
	      hitb->chid = OtrIdFromIndex(ci);
           }
	  }    
	  
	  // get measurement and error
	  float q,r ;
	  if(dglev_.idglev[LGESL_WCH-1] != 0) {    //long version
	    // Apply rt-relation:
	    q = dgocell::EstimatedRadius(digxIt->charge,gf->pit,gf->acc) ;
	    r = dgocell::EstimatedSigma(digxIt->charge,gf->pit,gf->acc) ;
	  } else {
	    q = digxIt->charge ;
	    r = gf->acc;
	  }
	  
	  // if no previous hit, take this digx; else take only if
	  // smaller drift distance.
	  if( prevstrip != digxIt->strip || q < hitb->q ) {
	    hitb->y  = digxIt->charge ;
	    hitb->q = q ;
	    hitb->r = r ;
	  }

	  // create relations
	  if(digxIt->mimp) 
	    createRelation(ArtePointer<MIMP>(digxIt->mimp),hitb,&MIMP::hitb);
	  createRelation(hitb,digb,&HITB::digb);

	  prevstrip = digxIt->strip ;
	}
      }     //======================== end of loop on digitized hits
   }        //======================== end of loop on sectors
   
   for(int i=0; i<4; i++)
     if(dg_otr[i]) { 
       delete dg_otr[i];  dg_otr[i] = 0 ;}
   
   ipl += nsec -1;
   
   OtrMcKillStatistics("Tstop", 0);
}




