#ifndef GROVER_WITHOUT_ARTE

#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"
#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// ----------------------------------------------------------------------------
// 00-09-20  Max      Skip very short MTRAs for improved speed
// 00-11-01  Max      Use cernlib random number gen instead of ROOT
// 00-11-03  Max      Error in 1/p covariance element found. The bug will
//                    only be present in ARTE-03-08-r2. Copy this file from
//                    r3 or later if you would like to use it. :-(
// 00-11-24  Ivan     Make loop over RSEGs instead of MTRAs
// 00-11-24  Ivan     Reinitialization now done by restoring initial cov matrix 
//                    from corresponding RPNTs
// 00-11-24  Max      Moved MTRA<->RSEG association to MTRAWhichCaused.C
// 00-11-26  Max      Added refit function for single RSEG + momentum + smearing value
// 00-12-01  Ivan     p_mc^-2  ->  (p_mc^2)/(p_smeared^4) for consistency
// 01-02-28  Max      const safe

extern "C" {
  void CatsRefit(ArtePointer<RSEG>);
  void rnorml_(float*, int*); // cernlib
}

void ReinitializeRSEG(ArtePointer<RSEG>);


//
// This could be streamlined to be more flexible
// - choice of all RSEGs or just one
// - user gives momentum value, else we find it in MTRA
// - user provides GroverCuts object (or smearing parameter)
//   else we just use defaults
// etc. etc.



int GroverRefitRSEG(const GroverCuts* usercuts = 0)
{

  /////////////////////////////////////////////////////////////////////
  //
  // Use momenta from MTRA to refit all VDS RSEGs using CatsRefit.
  // Momenta can be smeared by X% if desired by setting a cut.
  //
  /////////////////////////////////////////////////////////////////////


  bool debug = false;

  // error flag
  int errorflag = SUCCESS;

  // cuts
  static const GroverCuts temp;
  const GroverCuts* cuts = &temp;
  if (usercuts != 0) cuts = usercuts;
  // To smear or not to smear?
  bool  smearing    = (cuts->CheatSmearing == 1);
  // constant fractional error (if smearing)
  float smear_const =  cuts->CheatSmearConst;
  bool verbose = cuts->Verbose==1;

  if (!ArteTable<MCIN>::booked()) {
    if (verbose) 
      cout << "*** GroverRefitRSEG: WARNING - not Monte Carlo data!" <<endl;
    return FAILURE;
  }


  static bool firstcall = true;
  if (firstcall || verbose) {
    cout << "*** GroverRefitRSEG: refitting all RSEGs with MC momentum!" << endl;
    firstcall = false;
  }

  int idx = 0;

  // loop over vds rsegs and get the mtras corresponding to them
  for ( ArteTable<RSEG>::iterator rsegIt = ArteTable<RSEG>::begin();
       rsegIt != ArteTable<RSEG>::end(); ++rsegIt ) {
    if ( rsegIt->cmp  != (1 | Rsegc::bit16) ) continue; // not CATS

    ArtePointer<MTRA> mtra = MTRAWhichCaused(rsegIt);
    if (mtra != 0) {

      float mom = sqrt(mtra->px*mtra->px+mtra->py*mtra->py+mtra->pz*mtra->pz);
      float inv_mom = 1.0/mom;

      ReinitializeRSEG(*rsegIt); // restore initial cov matrix

      if (smearing) {
	//
	// (dp/p)^2 = constcon^2 + (p*varcon)^2
	//
	// and the TDR says:
	// ITR: constcon = 0.0040, varcon = 0.00002
	// OTR: constcon = 0.0050, varcon = 0.00005
	// (so roughly constant until high momenta, ~ 50 GeV)
	//
	// expected fractional error (squared)
	float avg_corr_sq =       smear_const*smear_const;
	// fractional errors for this track
	float constcon;
	int one=1;
	rnorml_(&constcon,&one);
	constcon = constcon*smear_const;

	// smear
	float smeared_mom = mom*(1.0+constcon);
	float smeared_inv_mom = 1.0/smeared_mom;
	(*rsegIt)->pf = (*rsegIt)->pe = smeared_inv_mom;
	// Variance of inverse momentum = (1/p^2 * fractional error)^2
	// = avg_corr_sq/p^2
	(*rsegIt)->cvf[14] = (*rsegIt)->cve[14] = 
	  (avg_corr_sq*mom*mom)/(smeared_mom*smeared_mom*smeared_mom*smeared_mom);
	// used to be (*rsegIt)->cvf[14] = (*rsegIt)->cve[14] = avg_corr_sq/(mom*mom);
	// All off-diagonals to zero
	for (int kk=10;kk<14;kk++) (*rsegIt)->cvf[kk] = 0.0;
	if (smeared_inv_mom <= 0.0) {
	  cout << "*** GroverRefitRSEG: ERROR - smeared momentum <= 0.0" << endl;
	  smeared_inv_mom = 0.0;
	  --errorflag;
	}
      }
      else { // not smearing - use momentum directly with small error
	(*rsegIt)->pf = (*rsegIt)->pe = inv_mom;
	(*rsegIt)->cvf[14] = (*rsegIt)->cve[14] = 0.00000001;
	for (int kk=10;kk<14;kk++) (*rsegIt)->cvf[kk] = 0.0;
      }
      CatsRefit(*rsegIt); // refit the track


    }
    if (debug) cout << endl; ++idx;
  }
  
  return errorflag;
}



//////////////////////////////////////////////////////////////////////////

int GroverRefitRSEG(ArtePointer<RSEG> rseg, const float p, const float smear = 0.0)
{
  // smear a single RSEG with the momentum given

  if (rseg == 0) return FAILURE;
  if (rseg->cmp  != (1|Rsegc::bit16)) return FAILURE; // not CATS
  if (p == 0.0) return FAILURE;

  bool smearing = (smear>0.0);
  if (smear > 0.5) return FAILURE; // ridiculous

  const float SMALLNUMBER = 0.00000001;

  float inv_mom = 1.0/p;
  ReinitializeRSEG(rseg); // restore initial cov matrix

  if (smearing) {
    // expected fractional error (squared)
    float avg_corr_sq =     smear*smear;
    // fractional errors for this track
    float constcon;
    int one=1;
    rnorml_(&constcon,&one); // cernlib
    constcon = constcon*smear;
    // smear
    float smeared_mom = p*(1.0+constcon);
    float smeared_inv_mom = 1.0/smeared_mom;
    if (smeared_inv_mom <= 0.0) {
      cout << "*** GroverRefitRSEG: ERROR - smeared momentum <= 0.0" << endl;
      return FAILURE;
    }
    rseg->pf = rseg->pe = smeared_inv_mom;
    // Variance of inverse momentum = (1/p^2 * fractional error)^2
    // = avg_corr_sq/p^2
    rseg->cvf[14] = rseg->cve[14] = 
      (avg_corr_sq*p*p)/(smeared_mom*smeared_mom*smeared_mom*smeared_mom);
    // used to be rseg->cvf[14] = rseg->cve[14] = avg_corr_sq/(p*p);
    // and all off-diagonals to zero
    for (int kk=10;kk<14;kk++) rseg->cvf[kk] = 0.0;
  }
  else { // not smearing - use momentum directly with small error
    rseg->pf = rseg->pe = inv_mom;
    rseg->cvf[14] = rseg->cve[14] = SMALLNUMBER;
    for (int kk=10;kk<14;kk++) rseg->cvf[kk] = 0.0;
  }



  CatsRefit(rseg); // refit the track
  return SUCCESS;
}


/////////////////////////////////////////////////////////////////////////


void ReinitializeRSEG(ArtePointer<RSEG> rseg) 
{ 
  //
  // from ARTE-03-08-r4 onwards, CATS stores the initial cov matrix in RPNT
  //

  int index = 0;

  for ( ArteRelationIterator<RPNT> rpntIt = rseg->rpnt.begin();
	rpntIt != rseg->rpnt.end(); ++rpntIt ){
    if (index == 0) for (int i=0; i<15; i++) rseg->cve[i] = (*rpntIt)->cvf[i];
    if (index == 1) for (int i=0; i<15; i++) rseg->cvf[i] = (*rpntIt)->cvf[i];
    ++index;
  }

}

#endif
