//--------------------------------------------------------------------
// Created: 17.03.99 by   Olya Igonkina
/* MARPLE function for matching RICH segments and ECAL clusters
   works both with magnet ON and OFF

   FORTRAN_ROUTINE(matchrichecal)( float * threshold_cut, 
                                   float * minRICHSB )
   if threshold_cut > 200 - it is Ktrig cut, otherwise it is pt_cut  
   minRICHSB is a signal/sqrt(background) ratio for RICH segments 
   saved as irseg->ze/sqrt(irseg->ye)

 */
//--------------------------------------------------------------------

#include "marple/marple.hh"
#include "marple/RcSegClusBuilder.hh"
#include "arte/RSEG_cmp.hh"

#include "marple/RcMcFunction.hh"

#ifdef MARPLEPERFORMANCES
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
Chrono * chMatchRichEcal = 0;
#endif


//definition of constants for matching ( see \Ref{RcMatchConst})
// flag, z_plane, z_min, z_max, use_pair_cut, n_pair_left, n_pair_right, 
//use_delta_cut, delta_x, delta_y, delta_tx, delta_ty, delta_pe, chi2

// to be optimized on MC !!!
// match with Ecluster  {tx, ty}  magnet is ON 
RcMatchConst rc2RichEcalConst = { 
  10,     1330.,   0.,  1500.,      0,          1,            2,
  1,     1.e+10  ,     1.e+10,   0.05,  0.0025,   10000.,   1.e+10
};
// match  SOFT
RcMatchConst rc2RichEcalSoftConst = { 
  10,     1330.,   0.,  1500.,      0,          2,            2,
  1,     1.e+10  ,     1.e+10,   0.05,  0.005,   10000.,   1.e+10
};


// match with Ecluster  {tx, ty}  magnet is OFF 
RcMatchConst rc2RichEcalNoMagnetConst = { 
  110,     1330.,   0.,  1500.,      1,          2,            2,
  2,     1.e+10  ,     1.e+10,   0.015,  0.016,   10000.,   140
};



extern "C"  
void FORTRAN_ROUTINE(matchrichecal)( float * threshold_cut,
				     float * minRICHSB) 
{
# ifdef MARPLEPERFORMANCES
  if(chMatchRichEcal == 0){
    ChronoManager CM;
    chMatchRichEcal = CM.createChrono("match Rich Ecal ");
  }
  chMatchRichEcal->start();
# endif

  RcSegClusBuilder RichEcalMatch;  // builder for charge track search
  maPresentMatch =MARPLEIDECALRICH;


  //--------------------------------------------------------------------
  // put constants to the MatchBuilder
  //--------------------------------------------------------------------
  if( MARPLE_SOFT )
    RichEcalMatch.setConst(     rc2RichEcalSoftConst );
  else
    {
      if( !maMagnetON() )
	RichEcalMatch.setConst(     rc2RichEcalNoMagnetConst );
      else
	RichEcalMatch.setConst(     rc2RichEcalConst );
    }
  RichEcalMatch.setBremConst( rc2RichEcalConst );

  //--------------------------------------------------------------------
  // create vector of RICH segments 
  //--------------------------------------------------------------------

  static std::vector<RcSegPoint * > nextRichSet;

  if( MARPLE_BKGR )
    {
      int foundRich = 0;
      for( ArteTable<RSEG>::iterator irseg = ArteTable<RSEG>::begin();
	   !foundRich && irseg != ArteTable<RSEG>::end(); ++irseg )
	if(  (*irseg)->cmp & Rsegc::rich  ) foundRich = 1;
      
      if( !foundRich )
	{
	  RichEcalMatch.deleteAll();  
	  
# ifdef MARPLEPERFORMANCES
	  chMatchRichEcal->stop();
# endif
	  maPresentMatch = MARPLENOMATCH;
	  return;
	}
      // segments from previous event
      for( unsigned int seg= 0; seg < nextRichSet.size(); seg++ )
	RichEcalMatch.addPointLeft( *(nextRichSet[seg]) );      
      nextRichSet.erase(nextRichSet.begin(), nextRichSet.end());
    }
 
  for( ArteTable<RSEG>::iterator irseg = ArteTable<RSEG>::begin();
       irseg != ArteTable<RSEG>::end(); ++irseg )
    {
      if( ! maRichSegment(*irseg) ) continue;
      if( (*irseg)->ye < MARPLE_NEGLIGIBLE ||
	  (*irseg)->ze/sqrt((*irseg)->ye)> *minRICHSB    )
	{ 	  
	  RcSegPoint * ptToPoint = new RcSegPoint( *irseg , 0); // firstPoint

	  if( MARPLE_BKGR )
	    nextRichSet.push_back(ptToPoint);
	  else
	    RichEcalMatch.addPointLeft( *ptToPoint);
	  ptToPoint = 0;
	}
    }
  if( RichEcalMatch.vectorLeft().size() == 0 )
    {
# ifdef MARPLEPERFORMANCES
      chMatchRichEcal->stop();
# endif
      maPresentMatch = MARPLENOMATCH;
      return;
    }

  //--------------------------------------------------------------------
  // create vector of clusters from CARE
  //--------------------------------------------------------------------

  int iErr  = 0;
  for( ArteTable<RCCL>::iterator irccl = ArteTable<RCCL>::begin();
       irccl != ArteTable<RCCL>::end(); ++irccl )
    {
      if( (*irccl)->flag < 0 ) continue;
      if( (*irccl)->esum < 1 ) continue;
      if( maTakeRccl(*threshold_cut, *irccl) )
	{
	  RcClusPoint * ptPoint = new RcClusPoint( (*irccl));
	  ptPoint->estimateDirectAngles(iErr);
	  RichEcalMatch.addPointRight( *ptPoint);	  
	  ptPoint = 0;
	}
    }  

  //--------------------------------------------------------------------
  // do Matching 
  //--------------------------------------------------------------------

  RichEcalMatch.matchRings();
  //--------------------------------------------------------------------
  // outprint
  //--------------------------------------------------------------------

  if( MARPLE_DEBUG >= 20 )
    {
      RichEcalMatch.fillMatch(8146);
      RichEcalMatch.fillDeltas(8246,8346);
    }else  if( MARPLE_DEBUG >= 10){
      RichEcalMatch.fillDeltas(8246);
    }

  RichEcalMatch.stat();
       
  
  //--------------------------------------------------------------------
  // save found relation to the table
  //--------------------------------------------------------------------
  if( MARPLE_MAKERSEG ) RichEcalMatch.saveRSEG(iErr);

  if( !MARPLE_BKGR ) 
    {
      RichEcalMatch.saveRelation(iErr);
      if( iErr ){ cout << "MARPLE: some relations were not saved " << endl;}
    }
  //--------------------------------------------------------------------
  // clean everything
  //--------------------------------------------------------------------
  RichEcalMatch.deleteAll();  
  maPresentMatch = MARPLENOMATCH;

# ifdef MARPLEPERFORMANCES
    chMatchRichEcal->stop();
# endif
}
