// 
//  Author:  Wouter Hulsbergen                 Created:  04-Jul-98
//  Email:   W.Hulsbergen@nikhef.nl            HERA-B Collaboration
// 
// Description: Set of global routines used to fill the hit patterns
// in RTRA/RTRE. Called from rangerTrack::storeback(..)
//
// Modifications:
//

#include <gean/ZGESL.h>
#include <arte/arte.hh>
#include <arte/RSEG_cmp.hh>
#include <digi/zgdig.hh>

#include <rang/Track.hh>
#include <rang/Node.hh>
#include <rang/Hit.hh>
#include <rang/HitPattern.hh>

#include <wax/DecodeManager.hh>
#include <Otr/OtrI.h>

extern int OtrEventChannelValid(int ci);

// This is a set of routines to fill the hit patterns in vds and
// otr. The inner tracker is not yet included.

namespace rangerHitPattern {

  //***********************************************************************
  //
  //  Routine:      prepareRTRE
  //                
  //  Author:       M. Dressel
  //
  //  Description:  
  //  Create entry in RTRE table for this rtra
  //
  //  Last modifications:
  //
  //***********************************************************************

  ArtePointer<RTRE> prepareRTRE(ArtePointer<RTRA> rtra)
  {
    // Create new entry in RTRE table
    int index = rtra.index() ;
    int pos   = ArteTable<RTRE>::size() ;
    if(pos < index ) {
      RTRE rtre ;
      rtre.mchi = rtre.pchi = rtre.tchi = 0 ;
      rtre.mche = rtre.pche = rtre.tche = 0 ;
      rtre.ze = rtre.xe = rtre.ye = rtre.txe = rtre.tye = rtre.pe = 0 ;
      ArteTable<RTRE>::allocate(index-pos) ;
      for(ArteTable<RTRE>::iterator irtre(pos+1) ; 
	  irtre != ArteTable<RTRE>::end() ; ++irtre) *(*irtre) = rtre ;
    }
    return ArtePointer<RTRE>(index) ;
  }

  //***********************************************************************
  //
  //  Routine:      SiPlaneIsHit
  //                
  //  Author:       M. Dressel
  //
  //  Description:  
  //  Calculate if vds plane is hit. 
  //
  //  Last modifications:
  //
  //***********************************************************************

  int SiPlaneIsHit( ArtePointer<GEDE> sigede, float xx, float yy){
    // SiPlaneIsHit checks if the track coordinates xx and yy
    // are in the sensitive area of the VDS GEDE plane sigede.
    //  
    // Input: all input is expected to be in global coordinates
    //        GEDE pointer to VDS plane: sigede 
    //        coordinates of a track on the GEDE plane: xx, yy
    //
    // return values: 
    // 1: VDS GEDE plane is crossed by extrapolated track
    // 0: VDS GEDE plane is not crossed.
    // negative values flag errors:
    // -1: unvalid GEDE plane
    
    const float LIMIT = 0.005; // to be 50 micron from the nominal boarders
    
    if ( sigede->cmp != 1) return -1; 
    
    int Nstrips = sigede->nwi[0];
    
    // sensitive area
    float sxmn = sigede->sxmn+LIMIT;
    float sxmx = sigede->sxmx-LIMIT;
    float symn = sigede->symn+LIMIT;
    float symx = sigede->symx-LIMIT;
    
    // check if track coordinates are on the sensitive area
    if ( xx < sxmn ||
	 xx > sxmx ||
	 yy < symn ||
	 yy > symx ) return 0;
    
    // projection of sensitive area on hitb->x coordinate
    float pit = sigede->pit;
    float xfw = sigede->xfw[0]; 
    float hbxmax = xfw - 0.5 * pit + pit*Nstrips   - LIMIT;
    float hbxmin = xfw - 0.5 * pit                 + LIMIT;
    
    // rotate positions of track
    float xx_rot = sigede->rot[0]*xx + sigede->rot[3]*yy;
    float yy_rot = sigede->rot[1]*xx + sigede->rot[4]*yy;
    
    // hitb->x of track coordinate
    float hbx = sigede->cos*xx_rot - sigede->sin*yy_rot;
    
    if ( hbx < hbxmin || hbx > hbxmax ) return 0;
    
    return 1;
    
  }

  //***********************************************************************
  //
  //  Routine:      fillVdsHitPattern
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill rtra->bse
  //
  //  Last modifications:
  //
  //***********************************************************************

  void fillVdsHitPattern(const rangerTrack* track, ArtePointer<RTRA> rtra)
  {
    typedef rangerTrack::rangerNodeConstIterator rangerNodeConstIterator ;
    unsigned int rc=0; 
    for(rangerNodeConstIterator nodeIt = track->begin() ;
	nodeIt != track->end() ; nodeIt++) {
      int plane = (*nodeIt)->plane() ;
      if( plane > 0 ) {
	ArtePointer<GEDE> gede = ArtePointer<GEDE>(plane) ;
	if(gede->cmp == LGESL_SVD) {
	  const rangerState& state = (*nodeIt)->getState() ;
	  float dz = 0.5*(gede->zmin+gede->zmax) - state.z() ;
	  float xx = state.x() + dz*state.tx() ;
	  float yy = state.y() + dz*state.ty() ;
	  // (bye the way, dz seems to be always zero)
	  if(  SiPlaneIsHit(gede,xx,yy)>0 ) {
	    int layer = 4 * (gede->gesl->nsl - 1) + gede->lay - 1;
	    rc |= 1 << layer ;
	  } 
	}
      }
    }
    rtra->bse = rc ;
  }
  
  //***********************************************************************
  //
  //  Routine:      fillOtrHitPattern
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill hit pattern entries in rtre
  //
  //  Last modifications:
  //
  //***********************************************************************

  unsigned char otrLayToBit( int geslnsl, int lay)
  {
    static unsigned char layindex[13][6]  ;
    static bool init=false ;
    if(!init) {
      init=true ;
      unsigned char nsinglelay[13][6] ;
      for( unsigned char isl=0; isl<13; isl++)
	for( unsigned char ilay=0; ilay<6; ilay++)
	  nsinglelay[isl][ilay]=0 ;
      unsigned char lastmc(0),lastpc(0) ;

      for(ArteTable<GESL>::iterator igesl = ArteTable<GESL>::begin() ;
	  igesl != ArteTable<GESL>::end() ; ++igesl) 
	if( igesl->cmp==2 ) {
	  for(ArteRelationIterator<GEDE> igede = (*igesl)->gede.begin() ;
	      igede != (*igesl)->gede.end() ; ++igede) 
	    if((*igede)->sec>0 &&
	       nsinglelay[igesl->nsl-1][(*igede)->lay-1]<(*igede)->nsl/2)
	      nsinglelay[igesl->nsl-1][(*igede)->lay-1]=(*igede)->nsl/2 ;
	  if( igesl->name[0] == 'M' ) lastmc = igesl->nsl-1 ;
	  if( igesl->name[0] == 'P' ) lastpc = igesl->nsl-1 ;
	}
      
      unsigned char thislayindex=0;
      for( unsigned char isl=0; isl<13; isl++) {
	for( unsigned char ilay=0; ilay<6; ilay++) {
	  layindex[isl][ilay]=thislayindex ;
	  thislayindex += nsinglelay[isl][ilay] ;
	}
	if( isl==lastmc || isl==lastpc ) thislayindex=0 ;
      }
    } 
    unsigned char rc =  layindex[geslnsl-1][lay-1] ;
    return rc ;
  }

  bool otrGedeWireIsGood(ArtePointer<GEDE> gede, int mono, int cell)
  {
    static int hitprepflag=-1 ;
    if( hitprepflag<0 ) {
      char* f ;
      if ((f= getenv("OtrHitPreparation"))) hitprepflag = atoi(f);
      else hitprepflag=0 ; //OTrI is default
    }
    bool rc=false ;
    if( hitprepflag==1 ) { // wax
      static waxDecodeManager* dm = waxDecodeManager::GetMe() ;
      const waxDecodeWire* wire = dm->GetGede(gede.index())->GetWire(cell+32*mono) ;
      if( wire && wire->IsConnected() && wire->GetChannel()->GetStatus()==0 )
	rc = true ;
    } else {               // OTrI
      int ci = OtrIFromNedeLayerWire(gede.index(),mono,cell) ;
      if( !OtrIError() && ci>0 && OtrEventChannelValid(ci)) rc=true ;
    } 
    return rc ;
  }

  void fillOtrHitPattern(const rangerTrack* track, ArtePointer<RTRE> rtre)
  {
    
    // calculates the number of crossed layers, and the number of found
    // hits in PC, MC, TC
    const float nsigmay = 2 ;
    const float nsigmax = 2 ;
    
    // reset
    rtre->mchi = rtre->pchi = rtre->tchi = 0 ;
    rtre->mche = rtre->pche = rtre->tche = 0 ;
    
    // loop over all gedes and see if they are really hit and if there is a hit
    ArtePointer<GEDE> gede ;
    for(rangerTrack::rangerNodeConstIterator inode = track->begin() ;
	inode != track->end() ; inode++) {
      int plane = (*inode)->plane() ;
      if(plane>0 && (gede=ArtePointer<GEDE>(plane))->cmp==2 &&
	 gede->sec>0 ) {
	float cos = gede->rot[0] ;
	float sin = gede->rot[1] ;
	const rangerState& state = (*inode)->getState() ;
	float z  = state.z() ; 
	float x  = state.x()*cos  - state.y()*sin ;
	float y  = state.y()*cos  + state.x()*sin ;
	float tx = state.tx()*cos - state.ty()*sin ;
	float ty = state.ty()*cos + state.tx()*sin ;
	const HepSymMatrix& cov = state.cov() ;
	float winy = nsigmay*sqrt(cos*cos*cov.fast(2,2) +
				  2*cos*sin*cov.fast(2,1) + sin*sin*cov.fast(1,1)) ;
	float winx = nsigmax*sqrt(cos*cos*cov.fast(1,1) -
				  2*cos*sin*cov.fast(2,1) + sin*sin*cov.fast(2,2)) ;
	
	float dzgede = 0.5*(gede->zmin+gede->zmax)-z ;
	
#ifdef FULL
	rangerVolume* vol = rangerGeom::vol(gede) ;
	float sLeng = vol->sLengNode(*inode) ;
	if(sLeng<=0) { 
	  gede->hwc[13]='\0' ;
	  cout << "volume no longer on track !!" << gede->zmin << " "
	       << (*inode)->hasHitb() << " " << gede->hwc << " " << plane << endl ;
	}
#endif	

	// first check in y en x of gede
	if(gede->symn + winy < y+dzgede*ty && y+dzgede*ty < gede->symx - winy &&
	   gede->sxmn + winx < x+dzgede*tx && x+dzgede*tx < gede->sxmx - winx ) {
	  
	  const float rmin = gede->pit/4 ; //--> closest cell to track
	  const float rmax = gede->pit/3 ;
	  
	  // are there any good cells on the track ? (within `small' window)
	  bool slhitexpected[2] = {false,false} ;
	  for(unsigned char mono=0; mono<gede->nsl; ++mono) {
	    // find closest cell in this mono layer
	    float thisx = x + (gede->zfw[mono]-z)*tx ;
	    // (the '3' is for safety: we don't want to truncate negative values)
	    unsigned char cell = int((thisx - gede->xfw[mono])/gede->pit+0.5+3) - 3 ;
	    if( cell<gede->nwi[mono] &&
		fabs( gede->xfw[mono] + cell*gede->pit - thisx ) < rmin &&
		otrGedeWireIsGood(gede,mono,cell) ) slhitexpected[mono/2] = true ;
	  }

	  // are there any hits on the track ? (within `large' window)
	  bool slhitfound[2]    = {false,false} ;
	  for(ArteTable<HITB>::iterator ihitb = gede->hit1 ;
	      ihitb != gede->hitl+1 ; ihitb++) 
	    if ( fabs(ihitb->x-(x + (ihitb->z-z)*tx) ) < rmax + winx ) {
	      int sl = 0;
	      if(gede->nsl>2 && ihitb->z > 0.5*(gede->zfw[1] + gede->zfw[2]) ) sl=1 ;
	      slhitfound[sl]=true ;
	    }
	  
	  for(unsigned char isl=0; isl<gede->nsl/2 ; ++isl) {
	    unsigned int layindex = otrLayToBit(gede->gesl->nsl,gede->lay) + isl ;
	    switch(gede->hwc[1]) {
	    case 'm':
	      if(slhitexpected[isl]) rtre->mche |= (1 << layindex) ;
	      if(slhitfound[isl])    rtre->mchi |= (1 << layindex) ;
	      break;
	    case 'p':
	      if(slhitexpected[isl]) rtre->pche |= (1 << layindex) ;
	      if(slhitfound[isl])    rtre->pchi |= (1 << layindex) ;
	      break;
	    case 't':
	      if(slhitexpected[isl]) rtre->tche |= (1 << layindex) ;
	      if(slhitfound[isl])    rtre->tchi |= (1 << layindex) ;
	      break;
	    }
	  }
	}
      }
    }
  }
  
  //***********************************************************************
  //
  //  Routine:      fillTrackEndPoint
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill state vector at last hit in rtre
  //
  //  Last modifications:
  //
  //***********************************************************************

  void fillTrackEndPoint(const rangerTrack* track, ArtePointer<RTRE> rtre)
  {
    for(rangerTrack::rangerNodeConstIterator nodeIt = track->begin() ;
	nodeIt != track->end() ; ++nodeIt) 
      if( (*nodeIt)->dim()>0 ) {
	const rangerState& state = (*nodeIt)->getState() ;
	rtre->ze  = state.z() ;
	rtre->xe  = state.x() ;
	rtre->ye  = state.y() ;
	rtre->txe = state.tx() ;
	rtre->tye = state.ty() ;
	rtre->pe  = state.qP() ;
	break;
      }
  }

  //***********************************************************************
  //
  //  Routine:      printHitPatterns(RTRA)   
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Print all hit patterns
  //
  //  Last modifications:
  //
  //***********************************************************************

  inline void printbitpattern(unsigned int hp)
  {
    for( int bit=0; bit<32; bit++)
      cout <<  (( hp & (1<<bit) ) ? '1' : '0') ;
    cout << endl ;
  }

  void printHitPatterns(ArtePointer<RTRA> rtra)
  {
    ArtePointer<RTRE> rtre(rtra.index()) ;
    cout << "hsi,hou,hca: " << rtra->hsi << " " 
	 << rtra->hou << " " << rtra->hca << endl ;
    cout << "rtra->bse:  " ; printbitpattern(rtra->bsi) ;
    cout << "rtra->bsi:  " ; printbitpattern(rtra->bse) ;
    cout << "rtre->mche: " ; printbitpattern(rtre->mche) ;
    cout << "rtre->mchi: " ; printbitpattern(rtre->mchi) ;
    cout << "rtre->pche: " ; printbitpattern(rtre->pche) ;
    cout << "rtre->pchi: " ; printbitpattern(rtre->pchi) ;
    cout << "rtre->tche: " ; printbitpattern(rtre->tche) ;
    cout << "rtre->tchi: " ; printbitpattern(rtre->tchi) ;
  }

  //***********************************************************************
  //
  //  Routine:      storeRtra    
  //                
  //  Author:       W. Hulsbergen
  //
  //  Description:  
  //  Fill rtra->bse and rtre table
  //
  //  Last modifications:
  //
  //***********************************************************************
  
  int storeRtra( rangerTrack* track, ArtePointer<RTRA> rtra)
  {  
    
    // lets' make this a uniform approach
    // prepare storing endpoint of track
    ArtePointer<RTRE> rtre = prepareRTRE(rtra) ;
 
    // 1) extrapolate track to vds and fill vds hit pattern
    if( rtra->cmp & Rsegc::vxd ) {
      track->extrapolateUpstream() ;
      fillVdsHitPattern(track,rtra) ;
    }
    
    // 2) extrapolate track to TC and fill otr hit pattern
    if( rtra->cmp & (Rsegc::patt | Rsegc::ecal) 
	/*&& (rtra->cmp & Rsegc::vxd)*/ ) {
      track->extrapolateDownstream() ;
      fillOtrHitPattern(track,rtre) ;
    }

    // 3) fill track parameters at last hit
    fillTrackEndPoint(track,rtre) ;
    
#ifdef FULL
    printHitPatterns(rtra) ;
#endif
    return 0;
  }

}
