// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

#include <iostream>
#include <algorithm>
#include "rang/VolTubs.hh"
#include "rang/FLAGRotFix.hh"
#include "rang/Arrow.hh"

// Before application of the rotation matrix, the tube/s axis is assumed
// to be identical with the z axis
// 
// The following information from GEDE/GWAL is used:
//
//   gwal->xmin   inner radius of tube
//       ->xmax   outer radius of tube
//       ->ymin   x offset of CM
//       ->ymax   y offset of CM
//       ->zmin   lower z bound
//       ->zmax   upper z bound
//       ->rot    rotation matrix
//       ->xmax   phi1
//       ->ymad   phi2
//
// See also gean/crvtrm.F


//-- Constructor
rangerVolTubs::rangerVolTubs(){};

//-- Constructor
rangerVolTubs::rangerVolTubs( rangerGeWa* gewa ) : rangerVolTube( gewa ) 
{
  // calculate the reference coordinates in ACS and PCS

  // this code is Rainer's: calculate central X and Y:
  double rMean = 0.5 * (Gewa->xmin()+Gewa->xmax());
  double phiMean = (Gewa->ymad() > Gewa->xmad())? 
    0.5 * (Gewa->xmad() + Gewa->ymad()) : 
    0.5 * (360 + Gewa->xmad() + Gewa->ymad());
  phiMean /= 57.295780; // convert to rad
  
  float zP = 0.5*(Gewa->zmin() + Gewa->zmax()) ;
  float xP = Gewa->ymin() + rMean * cos(phiMean) ;
  float yP = Gewa->ymax() + rMean * sin(phiMean) ;
  rangerArrow a(zP,xP,yP,1,1) ;
  rotateArrow(a,-1) ;
  zA = a.z ;

  // put box around in PCS and transform all 8 corners
  // we can make this box much smaller but it's okay for now
  xA1=yA1= 1.e9 ;
  xA2=yA2=-1.e9 ;
  float radius = max(Gewa->xmin(),Gewa->xmax()) ;
  for(int i=0; i<8; ++i) {
      xP  = Gewa->ymin() + ((i%4)/2 ? -1 :1)*radius ;
      yP  = Gewa->ymax() + ((i%2)   ? -1: 1)*radius ;
      zP = (i/4) ? Gewa->zmin() : Gewa->zmax() ;
      rangerArrow a(zP,xP,yP,1,1) ;
      rotateArrow(a,-1) ;
      xA1 = min(xA1,a.x) ;
      xA2 = max(xA2,a.x) ;
      yA1 = min(yA1,a.y) ;
      yA2 = max(yA2,a.y) ;
  }
}

//-- Destructor
rangerVolTubs::~rangerVolTubs() {};


bool rangerVolTubs::withinAngle(const rangerArrow& a, float z) const 
{
  double x = a.x + (z-a.z)*a.tx ;
  double y = a.y + (z-a.z)*a.ty ;
  double angle = atan2(y,x) * 57.295780;
  if (angle <0) angle += 360;
  double phi1 = Gewa->xmad();
  if (phi1<0) phi1 += 360;
  double phi2 = Gewa->ymad();
  if (phi2<0) phi2 += 360;
  return phi1 < angle && angle < phi2 ;
}


