// *****************************************************************************
//
// source:    rcClassRECDV.C
//
// type:      source code
//
// created:   10 Nov 2000
//	
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// symbol:    rcClassRECDV
//
// called by: rcClassify
//
// purpose:   check secondary vertex
//
// hints:     do NOT set flagdst here!
//
// changes:
// 10 Nov 2000 (AG) creation
// 14 Nov 2000 (AG) polished printout
// 30 Nov 2000 (SM) get the 2 leptons, and do the vertexing part
//  8 Dec 2000 (AG) moved globals for cuts to rcSelect
// 15 Dec 2000 (SM) totally changed: now the classification is done exclusively
//                  reading J/psi vertices in RVER. Detachment from closest wire
//  9 Jul 2001 (SM) modifications for third reprocessing:
//                  - use still the closest wire (not the primaries), but now chosen 
//                    with the distance from the J/psi mother track
//                  - detached vertices are chosen if they pass a cut on
//                    absolutle distance OR separation in sigma
//                  - the cuts set in the kumac are redefined: the first one stays
//                    the minimum separation in absolute distance, the second is
//                    redefined as minimum separation in sigma.
//                    The upper absolute separation is imposed by hand (3.5 cm)
// 10 Jul 2001 (SM) cleaning up, printouts removed 
//
// *****************************************************************************

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "grover/grover.hh"

// externals
extern int   FADEB;         // debug flag set by runargs=-debug
extern float cutRECDVlower; // lower cut
extern float cutRECDVupper; // upper cut

//
// function code
//
int
rcClassRECDV()
{
  
  // show up
  ARTE_mark( "rcClassRECDV 0", 0 );

  int rc = 0;
  if( FADEB ){
    cout << "RECDV: "
	 << "lower=" << cutRECDVlower << " "
	 << "upper=" << cutRECDVupper << " "
	 << endl;
  }

  float verdet = 0.;     // vertex detachment
  float verdets = 0.;    // vertex detachment in sigma
  float maxverdet = 0.;
  float maxverdets = 0.;
  
  // find in RVER all J/psi vertices 

  if (ArteTable<RVER>::size() !=0){
    for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
         rverIt != ArteTable<RVER>::end() ; ++rverIt){
      if( (*rverIt)->nt == 2   &&  ((*rverIt)->btra).index() != 0 ){
	if ( (*rverIt)->btra->fit == 2 && ((*rverIt)->btra->fid)%1000 == 443 ){

	  // FOR EVERY J/PSI CANDIDATE

	  // pick the closest wire

	  float mindist = 1e10; float dist = 1e9;
	  ArtePointer<RTAR> closest = 0;
	  for( ArteTable<RTAR>::iterator rtarget = ArteTable<RTAR>::begin();
	       rtarget!=ArteTable<RTAR>::end();
	       ++rtarget ) {
	    // consider only wires close to beam
	    if(sqrt(rtarget->x*rtarget->x+rtarget->y*rtarget->y)<1.0){ 


	      GroverParticle psi((*rverIt)->btra);
	      GroverWire tw(*rtarget);

	      // before we chose the wire closest in z to the J/psi vertex
	      // dist = fabs((*rverIt)->zv - (*rtarget)->z);
	      // now we pick up the one to which the J/psi mother track gets closest
	      dist = psi.DistanceTo(&tw);
	      
	      if(dist < mindist){
		mindist = dist;
		closest = (*rtarget);
	      }
	    }
	  }
	  
	  
	  float dzpsi = 0.; float dz = 0.;
	  if (closest != 0){
	    verdet = (*rverIt)->zv - closest->z;
	    dzpsi = ((*rverIt)->cv[5]>0.0) ? sqrt((*rverIt)->cv[5]) : 1e9;
	    dz = sqrt(dzpsi*dzpsi + ((closest->dz*closest->dz)/12));
	    verdets = verdet/dz;
	  }

	  // consider the largest separation in the event
	  if(fabs(verdet) > maxverdet){
	    maxverdet = fabs(verdet);
	    maxverdets = fabs(verdets);
	  }
	}
      }
    }

    if((fabs(maxverdet) >= cutRECDVlower ||  (fabs(maxverdets) >= cutRECDVupper))
       && fabs(maxverdet) <= 3.5){
      rc = 1;
    }
  }

  // show up
  ARTE_mark( "rcClassRECDV 1", 0 );
  
  return rc;
}
