// *****************************************************************************
//
// source:   Selection of Ds using the following channels:
//               Bs :  Ds- pi+, Ds- 2pi+ pi-
//               ~Bs:  Ds+ pi-, Ds+ 2pi- pi+
//            Returns:
//               1 upon success, 0 otherwise  
// dependences: 
//       - the following must be called BEFORE this routine:
//           selectDs()
//       - RTRA must be filled
//
//
// type:      source code, selection/reconstruction of Bs 
//
// created:   26 Aug 2000
//	
// author:   Denis Dujmic
//
// changes:
//    26 Aug 2000 (DD) >3.0 mm cut from a prim. vertex for testing
//    06 Sep 2000 (DD) File description added
//    15 DEc 2000 (DD) TOPP-RTRA relation required for Ds, remove MC code
//
// *****************************************************************************
#include "reco/rcX.h"
#include "reco/rcXkinema.h"
#include "rhp.h"
#include "dq/DQ.h"


#ifdef USE_ROOT_DD
#define MBS_LO 3.0
#define MBS_HI 10.0
#else
#define MBS_LO 4
#define MBS_HI 7
#endif

rcX *rcBs1=0;
rcX *rcBs2=0;
rcX *rcBs3=0;
rcX *rcBs4=0;

int Bs__Dsplus_Piminus_RC();
int Bs__Dsminus_Piplus_RC();
int Bs__Dsplus_Piminus_Piminus_Piplus_RC();
int Bs__Dsminus_Piminus_Piplus_Piplus_RC();

int maximizeRTRA();

int Bs() {
  Bs__Dsplus_Piminus_RC();
  Bs__Dsminus_Piplus_RC();
  Bs__Dsplus_Piminus_Piminus_Piplus_RC();
  Bs__Dsminus_Piminus_Piplus_Piplus_RC();    
  return 0;
}


/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  REAL RECO   /////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////




int Bs__Dsplus_Piminus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs1 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBs1) {
    cout<<"initializing Bs"<<endl;
    rcBs1=new rcX("Bs1", "Bs -> Ds+pi-", verb);
    decay bs;
    bs.fit         = 3;    // = short decay
    bs.book_tables = 1;    // 1/0
    bs.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    bs.verb        = verb;
    rcBs1->setMother(bs);  // set
    vertexAttrib vbs;
    vbs.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbs.topp               = kinema::findTopp(531); // pointer to mother
    vbs.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vbs.momentum_flag      = 1;     // momentum in vertex fit?
    vbs.check_wire         = 0;     // wire constraint?
    vbs.verb               = verb;  // verbosity 
    vbs.min_mass           = MBS_LO;// minimum mass
    vbs.max_mass           = MBS_HI;// maximum mass
    vbs.min_dz_from_wire   = .3;    // z-cut from primary vertex
    vbs.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBs1->setVertex(vbs);          // set
 
    attribute ds;
    ds.topp=kinema::findTopp(431);
    //ds.lund    = 431;
    //ds.mother  = 0;
    ds.matchTOPP =1;
    ds.fit     = 3;
    ds.mass    = ds.topp->mass;
    ds.minP    = 0;
    ds.maxP    = 200;
    ds.minMass = 1.9;
    ds.maxMass = 2.1;
    rcBs1->addDaughter(ds);
  
    attribute pi;
    pi.topp=kinema::findTopp(-211);
    //pi.lund         = -211;
    //pi.mother       = 531;
    pi.mass         = pi.topp->mass;
    pi.minP         = -200;
    pi.maxP         = 0;
    pi.minHIO       = 15;
    pi.minHSI       = 6;
    pi.minPt        = 0.15;
    pi.minLRPI      = .0001;    
    rcBs1->addDaughter(pi);

    if (verb) rcBs1->printDaughters();
  }
  rcBs1->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs1 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}

int Bs__Dsminus_Piplus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs2 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBs2) {
    cout<<"initializing Bs"<<endl;
    rcBs2=new rcX("Bs2", "Bs -> Ds-pi+", verb);
    decay bs;
    bs.fit        =3;    // = short decay
    bs.book_tables=1;    // 1/0
    bs.rlevel     =0;    // 0=full, 1=rtra, 2=mtra
    bs.verb       = verb;
    rcBs2->setMother(bs);// set
    vertexAttrib vbs;
    vbs.rlevel          = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbs.topp            = kinema::findTopp(531); // pointer to mother
    vbs.max_vertex_chi2 = 10.0;  // limit for vertex acceptance 
    vbs.momentum_flag   = 1;
    vbs.check_wire      = 0;
    vbs.verb            = verb;
    vbs.min_mass        = MBS_LO; // minimum mass
    vbs.max_mass        = MBS_HI; // maximum mass
    vbs.min_dz_from_wire   = .3;  // z-cut from primary vertex
    vbs.max_drho_from_wire = 0.02;// max impact parameter at primary vertex
    rcBs2->setVertex(vbs);        // set
 
    attribute ds;
    ds.topp=kinema::findTopp(-431);
    //ds.lund    = -431;
    //ds.mother  = 0;
    ds.fit     = 3;
    ds.matchTOPP =1;
    ds.mass    = ds.topp->mass;
    ds.minP    = -200;
    ds.maxP    = 0;
    ds.minMass = 1.9;
    ds.maxMass = 2.1;
    rcBs2->addDaughter(ds);
  
    attribute pi;
    pi.topp=kinema::findTopp(211);
    //pi.lund         = 211;
    //pi.mother       = 531;
    pi.mass         = pi.topp->mass;
    pi.minP         = 0;
    pi.maxP         = 200;
    pi.minHIO       = 15;
    pi.minHSI       = 6;
    pi.minPt        = 0.15;
    pi.minLRPI      = .0001;    
    rcBs2->addDaughter(pi);

    if (verb) rcBs2->printDaughters();
  }
  rcBs2->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs2 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


int Bs__Dsplus_Piminus_Piminus_Piplus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs3 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBs3) {
    cout<<"initializing Bs"<<endl;
    rcBs3=new rcX("Bs3", "Bs -> Ds+2pi-pi+", verb);
    decay bs;
    bs.fit        =3;     // = short decay
    bs.book_tables=1;     // 1/0
    bs.rlevel     =0;     // 0=full, 1=rtra, 2=mtra
    bs.verb       =verb;
    rcBs3->setMother(bs); // set
    vertexAttrib vbs;
    vbs.rlevel          = 2;      // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbs.topp            = kinema::findTopp(531); // pointer to mother
    vbs.max_vertex_chi2 = 10.0;   // limit for vertex acceptance 
    vbs.verb            = verb;
    vbs.check_wire      = 0;
    vbs.momentum_flag   = 1;      // momentum flag in mass fit
    vbs.min_mass        = MBS_LO; // minimum mass
    vbs.max_mass        = MBS_HI; // maximum mass
    vbs.min_dz_from_wire   = .3;  // z-cut from primary vertex
    vbs.max_drho_from_wire = 0.02;// max impact parameter at primary vertex
    rcBs3->setVertex(vbs);        // set
 
    attribute ds;
    ds.topp=kinema::findTopp(431);
    //ds.lund    = 431;
    //ds.mother  = 531;
    ds.fit     = 3;
    ds.matchTOPP=1;
    ds.mass    = ds.topp->mass;
    ds.minP    = 0;
    ds.maxP    = 200;
    ds.minMass = 1.9;
    ds.maxMass = 2.1;
    rcBs3->addDaughter(ds);
  
    attribute pi12;
    pi12.topp=kinema::findTopp(-211);
    //pi12.lund         = -211;
    //pi12.mother       = 531;
    pi12.mass         = pi12.topp->mass;
    pi12.minP         = -200;
    pi12.maxP         = 0;
    pi12.minHIO       = 15;
    pi12.minHSI       = 6;
    pi12.minPt        = 0.15;
    pi12.minLRPI      = .0001;    
    pi12.clone        = 2; // 2 x pi- 
    rcBs3->addDaughter(pi12);

    attribute pi3;
    pi3.topp=kinema::findTopp(211);
    //pi3.lund         = 211;
    //pi3.mother       = 531;
    pi3.mass         = pi3.topp->mass;
    pi3.minP         = 0;
    pi3.maxP         = 200;
    pi3.minHIO       = 15;
    pi3.minHSI       = 6;
    pi3.minPt        = 0.15;
    pi3.minLRPI      = .0001;    
    rcBs3->addDaughter(pi3);

    if (verb) rcBs3->printDaughters();
  }
  rcBs3->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs3 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


int Bs__Dsminus_Piminus_Piplus_Piplus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs4 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBs4) {
    cout<<"initializing Bs"<<endl;
    rcBs4=new rcX("Bs4", "Bs -> Ds-2pi+pi-", verb);
    decay bs;
    bs.fit        =3;     // = short decay
    bs.book_tables=1;     // 1/0
    bs.rlevel     =0;     // 0=full, 1=rtra, 2=mtra
    bs.verb       =verb;
    rcBs4->setMother(bs); // set
    vertexAttrib vbs;
    vbs.rlevel          = 2;      // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbs.topp            = kinema::findTopp(531); // pointer to mother
    vbs.max_vertex_chi2 = 10.0;   // limit for vertex acceptance 
    vbs.verb            = verb;
    vbs.check_wire      = 0;
    vbs.momentum_flag   = 1;
    vbs.min_mass        = MBS_LO;   // minimum mass
    vbs.max_mass        = MBS_HI;   // maximum mass
    vbs.min_dz_from_wire   = .3;    // z-cut from primary vertex
    vbs.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBs4->setVertex(vbs);          // set
 
    attribute ds;
    ds.topp=kinema::findTopp(-431);
    //ds.lund     = -431;
    //ds.mother   = 531;
    ds.fit      = 3;
    ds.matchTOPP=1;
    ds.mass     = ds.topp->mass;
    ds.minP     = -200;
    ds.maxP     = 0;
    ds.minMass  = 1.9;
    ds.maxMass  = 2.1;
    rcBs4->addDaughter(ds);
  
    attribute pi12;
    pi12.topp=kinema::findTopp(211);
    //pi12.lund         = 211;
    //pi12.mother       = 531;
    pi12.mass         = pi12.topp->mass;
    pi12.minP         = 0;
    pi12.maxP         = 200;
    pi12.minHIO       = 15;
    pi12.minHSI       = 6;
    pi12.minPt        = 0.15;
    pi12.minLRPI      = .0001;    
    pi12.clone        = 2; // 2 x pi+
    rcBs4->addDaughter(pi12);

    attribute pi3;
    pi3.topp  =kinema::findTopp(-211);
    //pi3.lund         = -211;
    //pi3.mother       = 531;
    pi3.mass         = pi3.topp->mass;
    pi3.minP         = -200;
    pi3.maxP         = 0;
    pi3.minHIO       = 15;
    pi3.minHSI       = 6;
    pi3.minPt        = 0.15;
    pi3.minLRPI      = .0001;    
    rcBs4->addDaughter(pi3);

    if (verb) rcBs4->printDaughters();
  }
  rcBs4->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Bs4 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}






//
//  Selection/reconstruction of Ds decays, returns 1 upon success
//
int
selectBs()
{
  ARTE_mark( "rcX_Bs 0", 0 );
  static int old_total1=0;
  static int old_total2=0;
  static int old_total3=0;
  static int old_total4=0;
  int ret=0;
  maximizeRTRA();
  rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1080 ), 0., 1.); // new event
  ArteTable<RTRA>::iterator b1=ArteTable<RTRA>::end();
  Bs();
  ArteTable<RTRA>::iterator b2=ArteTable<RTRA>::end();
  if (rcBs1->total>old_total1) { old_total1=rcBs1->total; ret=1; }
  if (rcBs2->total>old_total2) { old_total2=rcBs2->total; ret=1; }
  if (rcBs3->total>old_total3) { old_total3=rcBs3->total; ret=1; }
  if (rcBs4->total>old_total4) { old_total4=rcBs4->total; ret=1; }
  if (ret) {
    rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1080 ), 1., 1.); // Bs found
    // -- fill mass histogram
    for (ArteTable<RTRA>::iterator ib=b1; ib!=b2; ++ib) {
      ArtePointer<RVER> rver=(*ib)->ever;
      float mass=rver->mass;
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1081 ), mass, 1.);
    }
  }
  ARTE_mark( "rcX_Bs 1", 0 );
  return ret;
}



void stopBs() {
  if (rcBs1) delete rcBs1;
  if (rcBs2) delete rcBs2;
  if (rcBs3) delete rcBs3;
  if (rcBs4) delete rcBs4;
}
