
// *****************************************************************************
//
// source:    rcevnt.C
//
// type:      source code
//
// created:   17 Apr 1999
//	
// author:    Dr. Rainer Mankel
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Department of Physics
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// symbol:    rcevnt_
//
// called by: ATEXEC
//
// purpose:   reconstruction main program
//
// hints:     all reconstruction modules are called from here
//
// Recent Modifications:
// 18 Jan 2001 (GA) found timer bug
// 19 Jan 2001 (PC) added new scheme of RTRA creation
// 13 Jul 2001 (J.Hernandez) added new timer for RTRA
// 16 Jul 2001 (JH) added call to Cloneremove setting clone bit for RTRAs 
//  5 Oct 2001 (DS) vertexing is called also with RTRA2001 reconstruction scheme,
//                 optional call to the obsolete old complete 
//                 reconstruction scheme (rcZevnt) is removed
//  8 Oct 2001 (DS) removed call to v0k0s (obsolete due to new K0s finders)
//
// ******************************************************************************

// standard includes
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/ArteMessage.hh"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "marple/marple.hh"
#include "itrarte/ITR_arte.h"
#include "reco/rcflags.hh"
#include "reco/reco.hh"
#include "reco/eclass.hh"
#include "wax/OTrSelect.h"
#include "grover/grover.hh"

// globals
float time_rcMatch      = 0;
float time_rcRefit      = 0;
float time_rcParticleID = 0;
float time_rcVertex     = 0;
float time_rcRtra       = 0;

// externals
extern int         FATIME;        // set by runrags=-timing
extern int         FADEB;         // set by runargs=-debug
extern int         classify;
extern int         SLT_FILL;
extern int         acFlagReadITR;

// prototypes
void            SLTFillArteBanks(); // fill SLT-tables RSLA/RSLI
int             faCheckRhp();       // check for incoming messages
extern "C" void actaxxevnt_();      // TAXX
void            acLumiEvnt();       // HITL
int             l4tEvnt();          // L4 trigger
extern "C" void careinit_();        // CARE
extern "C" void tfinit_();          // Ranger
extern "C" void v0k0s_();           // K0 vertex
unsigned int    acFlag();
int             Cloneremove();      // clone bit

//
// program code
//
extern "C" void
rcevnt_()
{
  // mark
  ARTE_mark( "rcevnt BEGIN", 0 );

  //-----------------------------------------------------------------------------
  // initialization
  //-----------------------------------------------------------------------------

  // flag
  if( recoFlag ) rcSetError( "init" );
  
  // timers
  float time_rcEvnt           = 0;
  float time_rcInit           = 0;
  float time_rcStandAlone     = 0;
  float time_rcSelectTracks   = 0;
  float time_rcSetTrackPoints = 0;
  float time_rcSelectDecays   = 0;
  float time_rcClassify       = 0;
  float time_actaxxevnt       = 0;

  // flags
  int rcStandAloneFlag = 0;
  int matchingFlag     = 0;
  int RTRAflag         = 0;
  int refit            = 0;
  int particleID       = 0;
  int selectDecays     = 0;
  int momflag          = 0;
  int RPNTflag         = 0; 
  int l4tflag          = 0;

  //-----------------------------------------------------------------------------
  // create new EVRC bank if reco is switched on 
  //-----------------------------------------------------------------------------
  ArtePointer<EVRC> evrc;
  if( recoFlag ){
    int size = ArteTable<EVRC>::size();
    mmbook_( ArtePointer<EVRC>::getLink().getRef(), 1 );
    evrc = ArtePointer<EVRC>( size + 1 );
    evrc->ipaddr   = arteflag_.ipaddr;
    evrc->errmask  = 0x0;
    evrc->date     = time( NULL ); // now
    evrc->cputime  = 0.0;
    evrc->alca     = acFlag();
    evrc->artevers = arteflag_.artevers;
  }

  //-----------------------------------------------------------------------------
  // timing
  //-----------------------------------------------------------------------------
  static Chrono *cron_rcEvnt           = 0;
  static Chrono *cron_rcInit           = 0;
  static Chrono *cron_rcStandAlone     = 0;
  static Chrono *cron_rcMatch          = 0;
  static Chrono *cron_rcSelectTracks   = 0;
  static Chrono *cron_rcRefit          = 0;
  static Chrono *cron_rcSetTrackPoints = 0;
  static Chrono *cron_rcParticleID     = 0;
  static Chrono *cron_rcVertex         = 0;
  static Chrono *cron_rcSelectDecays   = 0;  
  static Chrono *cron_actaxxevnt       = 0;
  static Chrono *cron_rcClassify       = 0;
  static Chrono *cron_rcRtra           = 0;

  // initialize timers
  if( !cron_rcEvnt ){
    ChronoManager CM;
    CM.instance();

    cron_rcEvnt           = CM.createChrono( "rcEvnt" );
    cron_rcInit           = CM.createChrono( "rcInit" );
    cron_rcStandAlone     = CM.createChrono( "rcStandAlone" );
    cron_rcMatch          = CM.createChrono( "rcMatch" );
    cron_rcSelectTracks   = CM.createChrono( "rcSelectTracks" );
    cron_rcRefit          = CM.createChrono( "rcRefit" );
    cron_rcSetTrackPoints = CM.createChrono( "rcSetTrackPoints" );
    cron_rcParticleID     = CM.createChrono( "rcParticleID" );
    cron_rcVertex         = CM.createChrono( "rcVertex" );
    cron_rcSelectDecays   = CM.createChrono( "rcSelectDecays" );
    cron_actaxxevnt       = CM.createChrono( "actaxxevnt" ); 
    cron_rcClassify       = CM.createChrono( "rcClassify" ); 
    cron_rcRtra           = CM.createChrono( "rcRtra" ); 
  }

  // start timer
  cron_rcEvnt->start();

  //-----------------------------------------------------------------------------
  // SLT tables
  //-----------------------------------------------------------------------------
  if( SLT_FILL ){
    ARTE_mark( "SLTFillArteBanks", 0 );
    SLTFillArteBanks();
  }

  //-----------------------------------------------------------------------------
  // event reconstruction
  //-----------------------------------------------------------------------------
  if( recoFlag ){
    
    //==================================================================
    // 1. initialization, return for MC
    //==================================================================
    ArtePointer<EVHD> header( 1 );    
    if( ArteTable<MTRA>::begin() != ArteTable<MTRA>::end() ){
      if( header->exp > 0 ) header->exp = -1;
    }
    
    // rcinit
    static int first = 1;
    if( first ){
      first = 0; 

      // timer
      cron_rcInit->start();

      // ITR
      if( acFlagReadITR == 1 ){
	ARTE_mark( "ItrInit", 0 );
	ItrInit();
      }
      
      // ECAL
      ARTE_mark( "careinit", 0 );
      careinit_();
      
      // RANGER
      ARTE_mark( "tfinit", 0 );
      tfinit_();
      
      // MARPLE
      ARTE_mark( "marpleinit", 0 );
      marpleinit_();
      
      // timer
      cron_rcInit->stop();
      time_rcInit = cron_rcInit->lastTime();
      if( FATIME ) printf( "\033[1m TIME: init=%4.2f \033[0m \n", time_rcInit );
    }
    
    //==================================================================
    // stand alone reconstruction,
    // momentum determination of OTR segments
    //==================================================================

    cron_rcStandAlone->start();
    ARTE_mark( "rcStandAlone", 0 );
    rcStandAloneFlag = rcStandAlone();
    cron_rcStandAlone->stop();
    time_rcStandAlone = cron_rcStandAlone->lastTime();

    ARTE_mark( "rcMoment", 0 );
    momflag = rcMoment();

    //==================================================================
    // matching, track selection
    // If not rc2001: do standard chain: rcMatch, rcSelectTracks, refit.
    //==================================================================

    if(!rcRTRA2001){
      cron_rcMatch->start();
      ARTE_mark( "rcMatch", 0 );
      matchingFlag = rcMatch();
      cron_rcMatch->stop();
      time_rcMatch = cron_rcMatch->lastTime();

      cron_rcSelectTracks->start();
      ARTE_mark( "rcSelectTracks", 0 );
      RTRAflag = rcSelectTracks();
      cron_rcSelectTracks->stop();
      time_rcSelectTracks = cron_rcSelectTracks->lastTime();
    
      cron_rcRefit->start();
      ARTE_mark( "rcRefit", 0 );
      refit = rcRefit();
      cron_rcRefit->stop();
      time_rcRefit = cron_rcRefit->lastTime();

      cron_rcSetTrackPoints->start();
      ARTE_mark( "rcSetTrackPoints", 0 );
      RPNTflag = rcSetTrackPoints();
      cron_rcSetTrackPoints->stop();
      time_rcSetTrackPoints = cron_rcSetTrackPoints->lastTime();
    
    }
    else{

      //==================================================================
      // Follow new chain: 
      //==================================================================

      //starting timing
      cron_rcRtra->start();

      // Matching VDS-tracker
      ARTE_mark( "rcDoMatchStep", 0 );
      rcDoMatchStep(1); 
      
      //  create OTR-VDS and OTR standalone tracks and refit
      ARTE_mark( "rcCreateRtraBasePatt", 0 );
      rcCreateRtraBasePatt() ;
      ARTE_mark( "rcRefitRtraBasePatt", 0 );
      rcRefitRtraBasePatt() ;
      ARTE_mark( "rcSetTrackPoints", 0 );
      rcSetTrackPoints() ; // for tracks that were not refitted ?!
      
      // Selection of tracks vds-tracker:
      ARTE_mark( "rcSelectRtraBasePatt", 0 );
      rcSelectRtraBasePatt();

      // Step 2 of ECAL.
      if( recoECAL) rcEcalStep2();

      // Match 2:
      ARTE_mark( "rcDoMatchStep", 0 );
      rcDoMatchStep(2); 

      // rtra2:
      ARTE_mark( "rcCreateRtraLast", 0 );
      rcCreateRtraLast();
      ARTE_mark( "rcRefitRtraVdsEcal", 0 );
      rcRefitRtraVdsEcal() ;
      ARTE_mark( "rcRefitRtraVdsOnly", 0 );
      rcRefitRtraVdsOnly() ; // includes vds-rich
      ARTE_mark( "rcSetTrackPointsVdsEcal", 0 );
      rcSetTrackPointsVdsEcal() ;
      
      // assign RICH
      ARTE_mark( "rcDoMatchStep", 0 );
      rcDoMatchStep(3); 

      // match rtra to mtra
      ARTE_mark( "rcRTRAmatchMTRA", 0 );
      rcRTRAmatchMTRA() ;

      // stopping timing
      cron_rcRtra->stop();
      time_rcRtra = cron_rcRtra->lastTime();
    }

    //==================================================================
    // particle ID
    //==================================================================

    cron_rcParticleID->start();
    ARTE_mark( "rcParticleID", 0 );
    particleID = rcParticleID();
    cron_rcParticleID->stop();
    time_rcParticleID = cron_rcParticleID->lastTime();  
    
    //==================================================================
    // vertexing
    //==================================================================

    cron_rcVertex->start();
    ARTE_mark( "rcVertex", 0 );
    // vertex
    if( rcMainVertex == 1 || rcRTRA2001){
      ARTE_mark( "GroverEvent", 0 );
      GroverEvent();
    }
    // K0 vertex
//    if( rcK0SVertex == 1 ){
//      ARTE_mark( "v0k0s", 0 );
//      v0k0s_(); 
//    }
    cron_rcVertex->stop();
    time_rcVertex = cron_rcVertex->lastTime();
    
    //==================================================================
    // reconstruction of decays
    //==================================================================

    cron_rcSelectDecays->start();
    ARTE_mark( "rcSelectDecays", 0 );
    selectDecays = rcSelectDecays();
    cron_rcSelectDecays->stop();
    time_rcSelectDecays = cron_rcSelectDecays->lastTime();

    //==================================================================
    // fill EVHI bank
    //==================================================================

    ARTE_mark( "rcfillEVHI", 0 );
    rcfillEVHI();
    
    //==================================================================
    // print times
    //==================================================================

    if( FATIME ){      
      if (!rcRTRA2001) {
	float sum = 
	  time_rcStandAlone     +
	  time_rcMatch          +
	  time_rcSelectTracks   +
	  time_rcRefit          +
	  time_rcSetTrackPoints +
	  time_rcParticleID     +
	  time_rcVertex         +
	  time_rcSelectDecays;
      
	printf( "\033[1m TIME: s-a=%4.2f mtch=%4.2f trks=%4.2f ref=%4.2f pts=%4.2f pid=%4.2f vtx=%4.2f decs=%4.2f sum=%4.2f \033[0m \n",
		time_rcStandAlone, 
		time_rcMatch, 
		time_rcSelectTracks, 
		time_rcRefit, 
		time_rcSetTrackPoints, 
		time_rcParticleID, 
		time_rcVertex, 
		time_rcSelectDecays, 
		sum );
      } else {
	float sum = 
	  time_rcStandAlone     +
	  time_rcRtra           +
	  time_rcParticleID     +
	  time_rcVertex         +
	  time_rcSelectDecays;
      
	printf( "\033[1m TIME: s-a=%4.2f rtra2001=%4.2f pid=%4.2f vtx=%4.2f decs=%4.2f sum=%4.2f \033[0m \n",
		time_rcStandAlone, 
		time_rcRtra, 
		time_rcParticleID, 
		time_rcVertex, 
		time_rcSelectDecays, 
		sum );	
      }
    } 
  }

  //-----------------------------------------------------------------------------
  // track collection for alignment
  //-----------------------------------------------------------------------------

  cron_actaxxevnt->start();
  ARTE_mark( "actaxxevnt", 0 );
  actaxxevnt_();
  cron_actaxxevnt->stop();
  time_actaxxevnt = cron_actaxxevnt->lastTime();
  if( FATIME ) printf( "\033[1m TIME: taxx=%4.2f \033[0m \n", time_actaxxevnt );

  //-----------------------------------------------------------------------------
  // fill HITL for luminosity determination
  //-----------------------------------------------------------------------------

  ARTE_mark ( "acLumiEvnt", 0 );
  acLumiEvnt();

  //-----------------------------------------------------------------------------
  // event classification
  //-----------------------------------------------------------------------------

  if( classify ){
    cron_rcClassify->start();
    ARTE_mark( "rcClassify", 0 );
    rcClassify();
    cron_rcClassify->stop();
    time_rcClassify = cron_rcClassify->lastTime();    
    if( FATIME ) printf( "\033[1m TIME: class=%4.2f \033[0m \n", time_rcClassify );
  } 
  
  //-----------------------------------------------------------------------------
  // set flag correctly (NEEDED!)
  //-----------------------------------------------------------------------------

  ArtePointer<EVHD> evhd = *( ArteTable<EVHD>::end() - 1 );
  evhd->flag = evhd->flagdst;

  //==================================================================
  // Tag a clone-free subset out of the RTRA table
  //==================================================================

  if( recoFlag ){
    ARTE_mark( "Cloneremove", 0 );
    Cloneremove();
  }

  //-----------------------------------------------------------------------------
  // fill timing table EVTI
  //-----------------------------------------------------------------------------

  if( recoFlag ){
    ARTE_mark( "rcFillEVTI", 0 );
    rcFillEVTI();
  }

  //-----------------------------------------------------------------------------
  // stop overall timing
  //-----------------------------------------------------------------------------

  cron_rcEvnt->stop();
  time_rcEvnt = cron_rcEvnt->lastTime();

  //-----------------------------------------------------------------------------
  // fill in EVRC bank 
  //-----------------------------------------------------------------------------

  if( recoFlag ) {
    evrc->cputime = time_rcEvnt;
    evrc->errmask = recoErrorFlag();
    ARTE_mark( "OTrEventTime", 0 );
    OTrEventTime( evrc->ottmean,  evrc->ottwidth );
  }

  //-----------------------------------------------------------------------------
  // L4 trigger   
  //-----------------------------------------------------------------------------

  ARTE_mark( "l4tEvnt", 0 );
  l4tflag = l4tEvnt();

  //-----------------------------------------------------------------------------
  // return
  //-----------------------------------------------------------------------------

  ARTE_mark( "rcevnt END", 0 );
  return;
}
