// ---------------------------------------------------------------
//
//  update GEDE bank with values stored in database
//  and retrieved by LEDA
//
// ---------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "arte/arte.hh"
#include "silarte/silarte.hh"

#include "grr.h"

#include "aframelib.h"
#include "vdsleda.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

// array for faster retrieval of ARTE GEDE index from SL,lay,sec
static int gedeindex[8][4][4];

int silarte_updateGEDE()
{
  int  error = 0;

  char *routine = "silarte_updateGEDE";

  // loop over SIDE entries in LEDA
  for ( LedaTableIterator<class VDS_Side> side; side.valid(); side.next() ) {
    
    // loop over ARTE GEDE table to get corresponding GEDE row
    int found = 0;
    for ( ArteTable<GEDE>::iterator igede = ArteTable<GEDE>::begin();
	  igede != ArteTable<GEDE>::end(); igede++ ) {
      ArtePointer<GEDE> gede = *igede;

      if (   gede->cmp == 1
	  && strncmp( gede->gesl->name, side->slname, 4 ) == 0
	  && gede->lay == (int) side->lay
	  && gede->sec == (int) side->sec ) {

	// determine orientation of detector
	float deltax, deltay;
	if ( (int) side->sec == 2 || (int) side->sec == 4 ) {  // 3 and 9 o'clock
	  deltax = 5.;
	  deltay = 7.;
	}
	else {                                     // 6 and 12 o'clock
	  deltax = 7.;
	  deltay = 5.;
	}

	// update Arte GEDE table with SVD database values
	gede->pit    = side->pit;
	gede->cos    = cos(side->alpha * M_PI / 180.);
	gede->sin    = sin(side->alpha * M_PI / 180.);
	gede->xmin   = gede->xmin + side->sxmn          - gede->sxmn;
	gede->xmax   = gede->xmax + side->sxmn + deltax - gede->sxmx;
	gede->ymin   = gede->ymin + side->symn          - gede->symn;
	gede->ymax   = gede->ymax + side->symn + deltay - gede->symx;
	gede->zmin   = side->z - 0.0140;
	gede->zmax   = side->z + 0.0140; 
	gede->sxmn   = side->sxmn;
	gede->sxmx   = side->sxmn + deltax;
	gede->symn   = side->symn;
	gede->symx   = side->symn + deltay;
	gede->xfw[0] = gede->cos * gede->sxmn - gede->sin * gede->symn + side->d0;
	gede->xfw[1] = side->d0; // used in ARTE/alca/acGlobalAlignmentForGEDE.C
	gede->zfw[0] = side->z;

        // fill gede rotation matrix
        // NOTE: side->rot currently stores the inverse rotation!!!!!!!!!!
        // ====> a straight copy is not OK

        float da = vdsleda_get_zrot(side->rot)*M_PI/180.;
	for ( int i=0; i<9; i++ ) gede->rot[i] = 0; 
        gede->rot[8] = 1;
        gede->rot[0] =  gede->rot[4] = cos(da);
        gede->rot[1] =  sin(da);
        gede->rot[3] = -sin(da);

	// fill gedeindex array
	int sl = atoi( side->slname+2 );
	if (    sl              >= 1 && sl              <= 8
	     && (int) side->lay >= 1 && (int) side->lay <= 4
	     && (int) side->sec >= 1 && (int) side->sec <= 4 )
	  gedeindex[ sl-1 ][ (int) side->lay-1 ][ (int) side->sec-1 ] = gede.index();
	else
	  aframe_mess(routine,GRR_SWARNING,
		      "gedeindex array SL=%d, lay=%d, sec=%d ???",
		      sl, (int) side->lay, (int) side->sec );
	found = 1;

	// print out successful matching in debug mode
	if ( silarte_iol.debug_mode.value )
	  aframe_mess(routine,GRR_SINFO,
		      "Matched VDS side %d (SL=%s, lay=%d, sec=%d) to GEDE row %d.",
		      (int) side->id, side->slname, (int) side->lay, (int) side->sec, gede.index() );

      }
    }   // end loop over ARTE GEDE rows
    
    // complain in case no corresponding ARTE GEDE row was found
    if ( !found ) {
      aframe_mess(routine,GRR_SWARNING,
		  "VDS side %d (SL=%s, lay=%d, sec=%d) not found in GEDE table!",
		  (int) side->id, side->slname, (int) side->lay, (int) side->sec );
      error++;
    }

  }

  // Loop over ARTE GEDE tables to mark rows, which are not in SVD database
  for ( ArteTable<GEDE>::iterator igede = ArteTable<GEDE>::begin();
	igede != ArteTable<GEDE>::end(); igede++ ) {
    ArtePointer<GEDE> gede = *igede;
    if ( gede->cmp == 1 ) {

      int found = 0;
      for ( LedaTableIterator<class VDS_Side> side; side.valid(); side.next() )
	if (   strncmp( gede->gesl->name, side->slname,4 ) == 0
	    && gede->lay == (int) side->lay
	    && gede->sec == (int) side->sec )
	  found = 1;
      if ( !found ) {
	char geslname[5];
	strncpy( geslname,  gede->gesl->name, 4 );
	geslname[4] = '\0';
	aframe_mess(routine,GRR_SWARNING,
		    "GEDE row %d (SL=%s, lay=%d, sec=%d) not in VDS database", 
		    gede.index(), geslname, gede->lay, gede->sec);
      }
      
    }
  }
    
  if ( error ) 
    return GRR_FISHY;

  return GRR_OK;
}

//
//  return GEDE row from SL name, layer, and sector numbers
//
int silarte_returnGEDE( char* slname, int lay, int sec )
{
  int sl = atoi( slname+2 );
  if ( sl >= 1 && sl <= 8 && lay >= 1 && lay <= 4 && sec >= 1 && sec <= 4 )
    return gedeindex[ sl-1 ][ lay-1 ][ sec-1 ];
  else
    return 0;

}
